# """
# default parameters used to create realizations. This should be good for most applications
# """
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tidal_truncation import TruncationRN, TruncationRoche

class CosmoDefaults(object):

    def __init__(self):

        default_mass_function = 'sheth99'

        self.default_mass_function = default_mass_function

        # # default from WMAP9
        # self.H0 = 69.7
        # self.Ob0 = 0.0464
        # self.omega_DM = 0.235
        # self.Om0 = self.Ob0 + self.omega_DM
        # self.sigma8 = 0.82
        # self.curvature = 'flat'
        # self.ns = 0.9608
        # self.power_law = False

        # default from PLANCK2018
        self.H0 = 67.5
        self.Ob0 = 0.049
        self.omega_DM = 0.26
        self.Om0 = self.Ob0 + self.omega_DM
        self.sigma8 = 0.81
        self.curvature = 'flat'
        self.ns = 0.965
        self.power_law = False
        self.cosmo_param_dictionary = {'H0': self.H0, 'Ob0': self.Ob0, 'Om0': self.Om0,
                                        'Odm0': self.omega_DM, 'sigma8': self.sigma8, 'flat': self.curvature,
                                        'ns': self.ns, 'power_law': self.power_law}

    def __call__(self, key):

        try:
            return self.cosmo_param_dictionary[key]
        except:
            raise Exception(key + ' not a recognized cosmology key word argument.')


class LensConeDefaults(object):

    def __init__(self):
        self.default_zstart = 0.01
        self.distance_resolution_MPC = 1
        self.default_z_round = 2
        self.default_z_step = 0.02
        self.default_geometry = 'DOUBLE_CONE'
        self.kwargs_mass_sheet_default = {'kappa_scale': 1.0, 'log_mlow': 7.0,
                                          'log_mhigh': 10.0, 'subtract_exact_mass_sheets': False}
        # other possibilities:
        # CONE, CYLINDER

class TruncationDefaults(object):

    def __init__(self):

        self.RocheNorm = 1.4
        self.RocheNu = 2. / 3
        self.LOS_truncation = 50  # truncate at 'r50'
        self.truncation_class_subhalos = TruncationRoche(self.RocheNorm, RocheNu=self.RocheNu)
        self.truncation_class_field_halos = TruncationRN(self.LOS_truncation)

class DMHaloDefaults(object):

    def __init__(self):

        self.concentration_class_subhalos = ConcentrationDiemerJoyce
        self.concentration_class_field_halos = ConcentrationDiemerJoyce
        self.evaluate_mc_at_zlens = False

        self.scatter = True
        self.c_scatter_dex = 0.2

class RealizationDefaults(object):

    def __init__(self):

        # opening angle = opening_anlge_factor * Rein
        self.opening_angle_factor = 6

        self.default_r_tidal = '0.5Rs' # r_tidal = 'default_r_ridal * Rs'

        self.default_type = 'composite_powerlaw'

        self.default_mass_function = 'sheth99'

        self.default_subhalos_of_field_halos = False
        self.default_LOS_normalization = 1

        self.log_mlow = 6
        self.log_mhigh = 10

        self.host_m200 = 10 ** 13

        self.m_pivot = 10 ** 8

        self.delta_power_law_index = 0.
        self.delta_power_law_index_coupling = 1.

        self.subtract_exact_mass_sheets = False
        self.subhalo_mass_sheet_scale = 1.
        self.subtract_subhalo_mass_sheet = True
        self.draw_poisson = True

        self.subhalo_spatial_distribution = 'HOST_NFW'

        self.subhalo_convergence_correction_profile = 'NFW'

        self.kappa_scale = 1

        self.default_turnover_model = 'POLYNOMIAL'

####################################################################################

cosmo_default = CosmoDefaults()
lenscone_default = LensConeDefaults()
truncation_default = TruncationDefaults()
halo_default = DMHaloDefaults()
realization_default = RealizationDefaults()



import numpy.testing as npt
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Cosmology.cosmology import Cosmology
import astropy.units as un
import numpy as np
import pytest

class TestCosmology(object):

    def setup_method(self):

        self.arcsec = 2 * np.pi / 360 / 3600
        self.zlens = 1
        self.zsource = 2
        self.angle_diameter = 2/self.arcsec
        self.angle_radius = 0.5*self.angle_diameter

        self.H0 = 70
        self.omega_baryon = 0.03
        self.omega_DM = 0.25
        self.sigma8 = 0.82
        curvature = 'flat'
        self.ns = 0.9608
        cosmo_params = {'H0': self.H0, 'Om0': self.omega_baryon + self.omega_DM, 'Ob0': self.omega_baryon,
                      'sigma8': self.sigma8, 'ns': self.ns, 'curvature': curvature}
        self._dm, self._bar = self.omega_DM, self.omega_baryon
        self.cosmo = Cosmology(cosmo_kwargs=cosmo_params)
        self.geometry = Geometry(self.cosmo, self.zlens, self.zsource, self.angle_diameter, 'DOUBLE_CONE')

    def test_cosmo(self):

        da_true = self.cosmo.D_A(0, 1.824)
        da_interp = self.cosmo.D_A_z(1.824)
        npt.assert_almost_equal(da_true/da_interp, 1, 5)

        dc_true = self.cosmo.astropy.comoving_transverse_distance(1.4).value
        dc_interp = self.cosmo.D_C_z(1.4)
        dc_astropy = self.cosmo.astropy.comoving_distance(1.4).value
        npt.assert_almost_equal(dc_true/dc_interp, 1)
        npt.assert_almost_equal(dc_astropy/dc_true, 1)

        dc_transverse = self.cosmo.D_C_transverse(0.8)
        dc = self.cosmo.astropy.comoving_transverse_distance(0.8).value
        npt.assert_almost_equal(dc/dc_transverse, 1)

        ez = self.cosmo.E_z(0.8)
        ez_astropy = self.cosmo.astropy.efunc(0.8)
        npt.assert_almost_equal(ez / ez_astropy, 1)

        kpc_per_asec = self.cosmo.kpc_proper_per_asec(0.5)
        kpc_per_arcsec_true = self.cosmo.astropy.kpc_proper_per_arcmin(0.5).value / 60
        npt.assert_almost_equal(kpc_per_asec, kpc_per_arcsec_true, 2)

        rho_crit_0 = self.cosmo.rho_crit(0)
        rho_pc = un.Quantity(self.cosmo.astropy.critical_density(0), unit=un.Msun / un.pc ** 3)
        rho_Mpc = rho_pc.value * (1e+6) ** 3
        npt.assert_almost_equal(rho_crit_0/rho_Mpc, 1, 3)

        rho_crit = self.cosmo.rho_crit(0.6)
        rho_pc = un.Quantity(self.cosmo.astropy.critical_density(0.6), unit=un.Msun / un.m ** 3)
        rho_Mpc = rho_pc.value * self.cosmo.Mpc ** 3
        npt.assert_almost_equal(rho_crit / rho_Mpc, 1, 3)

        rho_crit_dark_matter = self.cosmo.rho_dark_matter_crit
        rho_crit_DM_astropy = self.cosmo.astropy.critical_density(0.).value * \
                              self.cosmo.density_to_MsunperMpc * self.cosmo.astropy.Odm(0.)
        npt.assert_almost_equal(rho_crit_DM_astropy, rho_crit_dark_matter)

        rho_crit = self.cosmo.rho_crit(0.3)
        rho_pc = un.Quantity(self.cosmo.astropy.critical_density(0.3), unit=un.Msun / un.pc ** 3)
        rho_Mpc = rho_pc.value * (1e+6) ** 3
        npt.assert_almost_equal(rho_crit / rho_Mpc, 1, 3)

        colossus = self.cosmo.colossus

        npt.assert_almost_equal(colossus.Om0, self.omega_DM + self.omega_baryon)
        npt.assert_almost_equal(colossus.Ob0, self.omega_baryon)
        npt.assert_almost_equal(colossus.H0, self.H0)
        npt.assert_almost_equal(colossus.sigma8, self.sigma8)
        npt.assert_almost_equal(colossus.ns, self.ns)

        halo_collapse_z = 9.
        age_today = self.cosmo.astropy.age(0.).value
        age_z = self.cosmo.astropy.age(halo_collapse_z).value
        age = age_today - age_z
        npt.assert_almost_equal(age, self.cosmo.halo_age(0., zform=halo_collapse_z))

        npt.assert_almost_equal(self.cosmo.scale_factor(0.7), self.cosmo.astropy.scale_factor(0.7))

if __name__ == '__main__':
     pytest.main()



from colossus.halo.concentration import *
from colossus.cosmology import cosmology
import astropy.cosmology as astropy_cosmo
from scipy.interpolate import interp1d
from pyHalo.defaults import *

cosmo_defaults = CosmoDefaults()

class Cosmology(object):

    M_sun = 1.98847 * 10 ** 30  # solar mass in [kg]

    Mpc = 3.08567758 * 10 ** 22  # Mpc in [m]

    arcsec = 2 * np.pi / 360 / 3600  # arc second in radian

    G = 6.67384 * 10 ** (-11) * Mpc**-3 * M_sun # Gravitational constant [Mpc^3 M_sun^-1 s^-2]

    c = 299792458*Mpc**-1 # speed of light [Mpc s^-1]

    density_to_MsunperMpc = 0.001 * M_sun**-1 * (100**3) * Mpc**3 # convert [g/cm^3] to [solarmasses / Mpc^3]

    def __init__(self, astropy_instance=None, cosmo_kwargs={}):

        self.astropy = self._setup_astropy_cosmology(astropy_instance, cosmo_kwargs)

        self._colossus_cosmo = self._setup_colossus_cosmology(cosmo_kwargs)

        self.h = self.astropy.h

        self._age_today = self.astropy.age(0).value

        self._DA_interp = self._interp_angular_diamter_distance()

        self._DC_interp = self._interp_comoving_distance()

        self._kpc_per_asec_interp = self._interp_kpc_per_asec()

    def D_A_z(self, z):

        try:
            return self._DA_interp(z)
        except:
            return self.D_A(0, z)

    def D_C_z(self, z):

        try:
            return self._DC_interp(z)
        except:
            return self.D_A_z(z) / self.scale_factor(z)

    def kpc_proper_per_asec(self, z):

        return self._kpc_per_asec_interp(z)

    @property
    def colossus(self):
        return self._colossus_cosmo

    def halo_age(self, z, zform=10):

        halo_form = self.astropy.age(zform).value
        if z > zform:
            return 0
        else:
            return self.astropy.age(z).value - halo_form

    def scale_factor(self, z):

        return (1 + z) ** -1

    def D_A(self, z1, z2):

        return self.astropy.angular_diameter_distance_z1z2(z1, z2).value

    def D_C_transverse(self, z):

        return self.astropy.comoving_transverse_distance(z).value

    def E_z(self, z):

        return self.astropy.efunc(z)

    def rho_crit(self, z):
        """
        :param z: redshift
        :return: critical density of the universe at redshift z in solar mass / Mpc^3
        """

        return self.astropy.critical_density(z).value * self.density_to_MsunperMpc

    @property
    def rho_dark_matter_crit(self):
        """
        :param z: redshift
        :return: comoving dark matter density of the universe at redshift z in solar mass / Mpc^3
        """

        return self.astropy.Odm(0.) * self.rho_crit(0.)

    def _interp_kpc_per_asec(self):

        zmin, zmax = 0.001, 4
        z = np.arange(zmin, zmax + 0.025, 0.025)
        kpc_per_asec = [self.astropy.arcsec_per_kpc_proper(zi).value ** -1 for zi in z]
        kpc_per_asec = np.array(kpc_per_asec)

        return interp1d(z, kpc_per_asec)

    def _interp_angular_diamter_distance(self):

        zmax = 4
        zstep = lenscone_default.default_z_step
        z = np.arange(zstep, zmax+zstep, zstep)
        da = []
        for zi in z:
            da.append(self.D_A(0, zi))
        da = np.array(da)
        return interp1d(z, da)

    def _interp_comoving_distance(self):

        zmax = 4
        zstep = lenscone_default.default_z_step
        z = np.arange(zstep, zmax + zstep, zstep)
        dc = []
        for zi in z:
            dc.append(self.D_C_transverse(zi))
        dc = np.array(dc)
        return interp1d(z, dc)

    def _setup_astropy_cosmology(self, astropy_instance, cosmo_kwargs):

        if not hasattr(self, 'astropy'):

            if astropy_instance is None:

                astropy_kwargs = {}

                keys = ['H0', 'Om0', 'Ob0']
                for key in keys:
                    if key not in cosmo_kwargs.keys():
                        astropy_kwargs.update({key: cosmo_defaults(key)})
                    else:
                        astropy_kwargs.update({key: cosmo_kwargs[key]})

                astropy_instance = astropy_cosmo.FlatLambdaCDM(**astropy_kwargs)

            self.astropy = astropy_instance

        return self.astropy

    def _setup_colossus_cosmology(self, cosmo_kwargs):

        if not hasattr(self,'colossus_cosmo'):

            colossus_kwargs = {}
            keys = ['H0', 'Om0', 'Ob0', 'ns', 'sigma8', 'power_law']
            for key in keys:
                if key not in cosmo_kwargs.keys():
                    colossus_kwargs.update({key: cosmo_defaults(key)})
                else:
                    colossus_kwargs.update({key: cosmo_kwargs[key]})
            self._colossus_cosmo = cosmology.setCosmology('custom', colossus_kwargs)

        return self._colossus_cosmo



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa
import numpy as np

class Gaussian(Halo):
    """
    The base class for a Gaussian fluctuation

    # kappa0 = amp / (2 * np.pi * sigma ** 2)
    """

    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        mdef = 'GAUSSIAN_KAPPA'
        super(Gaussian, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag, fixed_position=True)

    @property
    def profile_args(self):
        return None

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['GAUSSIAN_KAPPA']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        kwargs=[{'amp':self._args['amp'],
                'sigma':self._args['sigma'],
                'center_x':self._args['center_x'],
                'center_y':self._args['center_y']}]
        return kwargs, None

    @property
    def z_eval(self):
        """
        Returns the halo redshift
        """
        return self.z



import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.interpolate import interp1d
from lenstronomy.LensModel.Profiles.tnfw import TNFW

def tnfw_density_profile(r, rhos, rs, rt):
    """

    :param r:
    :param rhos:
    :param rs:
    :param r_t:
    :return:
    """
    x = r/rs
    tau = rt/rs
    nfw_prof = rhos / (x * (1+x)**2)
    truncation = tau ** 2 / (x**2 + tau**2)
    return nfw_prof * truncation

def tnfw_mass_fraction(tau, c):
    """
    This returns the ratio of the final mass to the initial mass of a truncated NFW profile given the truncation radius
    in units of rs (tau) and the concentration (c)
    :param tau: halo truncation radius in units of the scale radius
    :param c: the halo concentration
    :return: the ratio m_bound / m_virial
    """
    prof = TNFW()
    rs = 1.0
    rho0 = 1 / (4 * np.pi * rs ** 3)
    m = prof.mass_3d(c, rs, rho0, tau)
    nfw_func = np.log(1 + c) - c/(1+c)
    return m/nfw_func

def tau_mf_interpolation():

    """
    This function interpolates solutions for the truncation radius of a truncated NFW profile given the concentration
    and final bound mass
    :return: an instance of RegularGridInterpolator that returns (r_t / r_s) given a final mass and concentration
    """
    N = 50
    tau = np.logspace(-1., 2., N)
    log10_c = np.linspace(0, 2.7, N)
    # mass_fraction_1d = np.logspace(-1.45, -0.02, N)
    mass_fraction_1d = np.logspace(-1.5, np.log10(0.999), N)
    log10tau_2d = np.zeros((N, N))

    # This computes the value of tau that correponds to each pair of (concentration, mass_loss)
    for i, log10con_i in enumerate(log10_c):
        log10final_mass = np.log10(tnfw_mass_fraction(tau, 10**log10con_i))
        mfinterp = interp1d(log10final_mass, np.log10(tau), bounds_error=False, fill_value='extrapolate')
        for j, mass_j in enumerate(mass_fraction_1d):
            log10tau_2d[i, j] = mfinterp(np.log10(mass_j))

    interp_points = (log10_c, np.log10(mass_fraction_1d))
    interpolator = RegularGridInterpolator(interp_points, log10tau_2d, bounds_error=False, fill_value=None)
    return interpolator



"""
These functions define some preset halo mass function models that can be specified instead of individually specifying
a bunch of keyword arguments. This is intended for a quick and easy user interface. The definitions shown here can also
act as a how-to guide if one wants to explore more complicated descriptions of the halo mass function, as the models
presented here show what each keyword argument accepted by pyHalo does.
"""
from pyHalo.pyhalo import pyHalo
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw
from pyHalo.Halos.HaloModels.TNFWemulator import TNFWSubhaloEmulator
from pyHalo.Rendering.MassFunctions.density_peaks import ShethTormen
from pyHalo.Rendering.SpatialDistributions.uniform import LensConeUniform
from pyHalo.Rendering.SpatialDistributions.nfw import ProjectedNFW
from pyHalo.truncation_models import truncation_models
from pyHalo.concentration_models import preset_concentration_models
from pyHalo.mass_function_models import preset_mass_function_models
from pyHalo.single_realization import Realization
from copy import copy
import numpy as np
from pyHalo.realization_extensions import RealizationExtensions
from pyHalo.utilities import de_broglie_wavelength, MinHaloMassULDM

__all__ = ['preset_model_from_name', 'CDM', 'WDM', 'ULDM', 'SIDM_core_collapse', 'WDM_mixed',
           'CDMFromEmulator', 'WDMGeneral']

def preset_model_from_name(name):
    """
    Retruns a preset_model function from a string
    :param name: the name of the preset model, should be the name of a function in this file
    :return: the function
    """
    if name == 'CDM':
        return CDM
    elif name == 'WDM':
        return WDM
    elif name == 'SIDM_core_collapse':
        return SIDM_core_collapse
    elif name == 'ULDM':
        return ULDM
    elif name == 'CDMEmulator':
        return CDMFromEmulator
    elif name == 'WDM_mixed':
        return WDM_mixed
    elif name == 'WDMGeneral':
        return WDMGeneral
    else:
        raise Exception('preset model '+ str(name)+' not recognized!')

def CDM(z_lens, z_source, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_ROCHE_GILMAN2020', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, two_halo_contribution=True, delta_power_law_index=0.0,
        geometry_type='DOUBLE_CONE', kwargs_cosmo=None):
    """
    This class generates realizations of dark matter structure in Cold Dark Matter

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param LOS_normalization: rescales the amplitude of the line-of-sight halo mass function
    :param two_halo_contribution: bool; turns on and off the two-halo term
    :param delta_power_law_index: tilts the logarithmic slope of the subhalo and field halo mass functions around pivot
    at 10^8 M_sun
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology (see pyHalo.Cosmology.cosmology)
    :return: a realization of dark matter halos
    """

    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # NOW WE SET THE MASS FUNCTION CLASSES FOR SUBHALOS AND FIELD HALOS
    # NOTE: MASS FUNCTION CLASSES SHOULD NOT BE INSTANTIATED HERE
    mass_function_model_subhalos = CDMPowerLaw
    mass_function_model_fieldhalos = ShethTormen

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    subhalo_spatial_distribution = ProjectedNFW
    fieldhalo_spatial_distribution = LensConeUniform

    # set the density profile definition
    mdef_subhalos = 'TNFW'
    mdef_field_halos = 'TNFW'

    kwargs_concentration_model_subhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_fieldhalos['cosmo'] = pyhalo.astropy_cosmo

    # SET THE CONCENTRATION-MASS RELATION FOR SUBHALOS AND FIELD HALOS
    model_subhalos, kwargs_mc_subs = preset_concentration_models(concentration_model_subhalos,
                                                                 kwargs_concentration_model_subhalos)
    concentration_model_subhalos = model_subhalos(**kwargs_mc_subs)

    model_fieldhalos, kwargs_mc_field = preset_concentration_models(concentration_model_fieldhalos,
                                                                    kwargs_concentration_model_fieldhalos)
    concentration_model_fieldhalos = model_fieldhalos(**kwargs_mc_field)

    # SET THE TRUNCATION RADIUS FOR SUBHALOS AND FIELD HALOS
    kwargs_truncation_model_subhalos['lens_cosmo'] = pyhalo.lens_cosmo
    kwargs_truncation_model_fieldhalos['lens_cosmo'] = pyhalo.lens_cosmo

    model_subhalos, kwargs_trunc_subs = truncation_models(truncation_model_subhalos)
    kwargs_trunc_subs.update(kwargs_truncation_model_subhalos)
    truncation_model_subhalos = model_subhalos(**kwargs_trunc_subs)

    model_fieldhalos, kwargs_trunc_field = truncation_models(truncation_model_fieldhalos)
    kwargs_trunc_field.update(kwargs_truncation_model_fieldhalos)
    truncation_model_fieldhalos = model_fieldhalos(**kwargs_trunc_field)

    # NOW THAT THE CLASSES ARE SPECIFIED, WE SORT THE KEYWORD ARGUMENTS AND CLASSES INTO LISTS
    population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT']
    mass_function_class_list = [mass_function_model_subhalos, mass_function_model_fieldhalos]
    kwargs_subhalos = {'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                       'm_pivot': 10 ** 8,
                       'power_law_index': shmf_log_slope,
                       'delta_power_law_index': delta_power_law_index,
                       'log_m_host': log_m_host,
                       'sigma_sub': sigma_sub}
    kwargs_los = {'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                       'm_pivot': 10 ** 8,
                       'delta_power_law_index': delta_power_law_index,
                       'log_m_host': log_m_host}
    kwargs_mass_function_list = [kwargs_subhalos, kwargs_los]
    spatial_distribution_class_list = [subhalo_spatial_distribution, fieldhalo_spatial_distribution]
    kwargs_subhalos_spatial = {'m_host': 10 ** log_m_host, 'zlens': z_lens,
                               'rmax2d_arcsec': cone_opening_angle_arcsec / 2, 'r_core_units_rs': r_tidal,
                               'lens_cosmo': pyhalo.lens_cosmo}
    kwargs_los_spatial = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}
    kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial]

    if two_halo_contribution:
        population_model_list += ['TWO_HALO']
        kwargs_two_halo = copy(kwargs_los)
        kwargs_mass_function_list += [kwargs_two_halo]
        spatial_distribution_class_list += [LensConeUniform]
        kwargs_spatial_distribution_list += [kwargs_los_spatial]
        mass_function_class_list += [ShethTormen]

    kwargs_halo_model = {'truncation_model_subhalos': truncation_model_subhalos,
                         'concentration_model_subhalos': concentration_model_subhalos,
                         'truncation_model_field_halos': truncation_model_fieldhalos,
                         'concentration_model_field_halos': concentration_model_fieldhalos,
                         'kwargs_density_profile': {}}

    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                          spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                          geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model, nrealizations=1)
    return realization_list[0]

def WDM(z_lens, z_source, log_mc, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        mass_function_model_subhalos='LOVELL2020', kwargs_mass_function_subhalos={},
        mass_function_model_fieldhalos='LOVELL2020', kwargs_mass_function_fieldhalos={},
        concentration_model_subhalos='BOSE2016', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='BOSE2016', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_ROCHE_GILMAN2020', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        mdef_subhalos='TNFW', mdef_field_halos='TNFW', kwargs_density_profile={}):

    """
    This class generates realizations of dark matter structure in Warm Dark Matter

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param log_mc: the log base 10 of the half-mode mass
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param mass_function_model_subhalos: mass function model for subhalos, see mass_function_models.py for a list
    :param kwargs_mass_function_subhalos: keyword arguments for the mass function model
    :param mass_function_model_fieldhalos: mass function model for field halos, see mass_function_models.py for a list
    :param kwargs_mass_function_fieldhalos: keyword arguments for the mass function model
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology (see pyHalo.Cosmology.cosmology)
    :param mdef_subhalos: mass definition for subhalos
    :param mdef_field_halos: mass definition for field halos
    :param kwargs_density_profile: keyword arguments for the specified mass profile
    :return: a realization of dark matter halos
    """
    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    subhalo_spatial_distribution = ProjectedNFW
    fieldhalo_spatial_distribution = LensConeUniform

    # SET THE MASS FUNCTION MODELS FOR SUBHALOS AND FIELD HALOS
    # NOTE: MASS FUNCTIONS SHOULD NOT BE INSTANTIATED HERE
    mass_function_model_subhalos, kwargs_mass_function_subhalos = preset_mass_function_models(mass_function_model_subhalos,
                                                                                              kwargs_mass_function_subhalos)
    kwargs_mass_function_subhalos['log_mc'] = log_mc

    mass_function_model_fieldhalos, kwargs_mass_function_fieldhalos = preset_mass_function_models(mass_function_model_fieldhalos,
                                                                                                  kwargs_mass_function_fieldhalos)
    kwargs_mass_function_fieldhalos['log_mc'] = log_mc

    # SET THE CONCENTRATION-MASS RELATION FOR SUBHALOS AND FIELD HALOS
    kwargs_concentration_model_subhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_subhalos['log_mc'] = log_mc
    kwargs_concentration_model_fieldhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_fieldhalos['log_mc'] = log_mc

    model_subhalos, kwargs_mc_subs = preset_concentration_models(concentration_model_subhalos,
                                                                 kwargs_concentration_model_subhalos)
    concentration_model_CDM = preset_concentration_models('DIEMERJOYCE19')[0]
    kwargs_mc_subs['concentration_cdm_class'] = concentration_model_CDM
    concentration_model_subhalos = model_subhalos(**kwargs_mc_subs)

    model_fieldhalos, kwargs_mc_field = preset_concentration_models(concentration_model_fieldhalos,
                                                                    kwargs_concentration_model_fieldhalos)
    kwargs_mc_field['cosmo'] = pyhalo.astropy_cosmo
    kwargs_mc_field['log_mc'] = log_mc
    concentration_model_CDM = preset_concentration_models('DIEMERJOYCE19')[0]
    kwargs_mc_field['concentration_cdm_class'] = concentration_model_CDM
    concentration_model_fieldhalos = model_fieldhalos(**kwargs_mc_field)

    # SET THE TRUNCATION RADIUS FOR SUBHALOS AND FIELD HALOS
    kwargs_truncation_model_subhalos['lens_cosmo'] = pyhalo.lens_cosmo
    kwargs_truncation_model_fieldhalos['lens_cosmo'] = pyhalo.lens_cosmo

    model_subhalos, kwargs_trunc_subs = truncation_models(truncation_model_subhalos)
    kwargs_trunc_subs.update(kwargs_truncation_model_subhalos)
    kwargs_trunc_subs['lens_cosmo'] = pyhalo.lens_cosmo
    truncation_model_subhalos = model_subhalos(**kwargs_trunc_subs)

    model_fieldhalos, kwargs_trunc_field = truncation_models(truncation_model_fieldhalos)
    kwargs_trunc_field.update(kwargs_truncation_model_fieldhalos)
    kwargs_trunc_field['lens_cosmo'] = pyhalo.lens_cosmo
    truncation_model_fieldhalos = model_fieldhalos(**kwargs_trunc_field)

    # NOW THAT THE CLASSES ARE SPECIFIED, WE SORT THE KEYWORD ARGUMENTS AND CLASSES INTO LISTS
    population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT', 'TWO_HALO']

    mass_function_class_list = [mass_function_model_subhalos,
                                mass_function_model_fieldhalos,
                                mass_function_model_fieldhalos]
    kwargs_mass_function_subhalos.update({'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                       'm_pivot': 10 ** 8,
                       'power_law_index': shmf_log_slope,
                       'log_m_host': log_m_host,
                       'sigma_sub': sigma_sub,
                       'delta_power_law_index': 0.0})
    kwargs_mass_function_fieldhalos.update({'log_mlow': log_mlow,
                  'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                  'm_pivot': 10 ** 8,
                  'log_m_host': log_m_host,
                  'delta_power_law_index': 0.0})
    kwargs_mass_function_list = [kwargs_mass_function_subhalos, kwargs_mass_function_fieldhalos, kwargs_mass_function_fieldhalos]
    spatial_distribution_class_list = [subhalo_spatial_distribution, fieldhalo_spatial_distribution, fieldhalo_spatial_distribution]
    kwargs_subhalos_spatial = {'m_host': 10 ** log_m_host, 'zlens': z_lens,
                               'rmax2d_arcsec': cone_opening_angle_arcsec / 2, 'r_core_units_rs': r_tidal,
                               'lens_cosmo': pyhalo.lens_cosmo}
    kwargs_los_spatial = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}
    kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial, kwargs_los_spatial]

    kwargs_halo_model = {'truncation_model_subhalos': truncation_model_subhalos,
                         'concentration_model_subhalos': concentration_model_subhalos,
                         'truncation_model_field_halos': truncation_model_fieldhalos,
                         'concentration_model_field_halos': concentration_model_fieldhalos,
                         'kwargs_density_profile': kwargs_density_profile}

    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                     spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                     geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model, nrealizations=1)
    return realization_list[0]

def ULDM(z_lens, z_source, log10_m_uldm, log10_fluc_amplitude=-0.8, fluctuation_size_scale=0.05,
          fluctuation_size_dispersion=0.2, n_fluc_scale=1.0, velocity_scale=200, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        mass_function_model_subhalos='SHMF_SCHIVE2016', kwargs_mass_function_subhalos={},
        mass_function_model_fieldhalos='SCHIVE2016', kwargs_mass_function_fieldhalos={},
        concentration_model_subhalos='LAROCHE2022', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='LAROCHE2022', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_ROCHE_GILMAN2020', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
         uldm_plaw=1 / 3, flucs=True, flucs_shape='aperture', flucs_args={}, n_cut=50000, r_ein=1.0):

    """
    This generates realizations of ultra-light dark matter (ULDM), including the ULDM halo mass function and halo density profiles,
    as well as density fluctuations in the main deflector halo.

    Similarly to WDMGeneral, the functional form of the subhalo mass function is the same as the field halo mass function.
    However, this model differs from WDMGeneral by creating halos which are composite ULDM + NFW density profiles.
    The ULDM particle mass and core radius-halo mass power law exponent must now be specified. For details regarding ULDM halos,
    see Schive et al. 2014 (https://arxiv.org/pdf/1407.7762.pdf). Equations (3) and (7) give the soliton density profile
    and core radius, respectively.

    The differential halo mass function is described by three parameters, see Schive et al. 2016
    (https://arxiv.org/pdf/1508.04621.pdf):

    1) log_m0 - the log10 value of the characteristic mass, or the scale where the ULDM mass function begins
    to deviate from CDM:

        m0 = (1.6*10**10) * (m22)**(-4/3)   [solar masses],

    where

        m22 = m_uldm / 10**22 eV.

    2) b_uldm - modifies the log slope of the ULDM mass function, analogous to b_wdm (see WDMLovell2020)

    3) c_uldm - modifies the log slope of the ULDM mass function, analogous to c_wdm (see WDMLovell2020)

    The parametrization for the mass function is:

        n_uldm / n_cdm = (1 + (m / m_0)^b_uldm) ^ c_uldm,

    where Schive et al. 2016 determined that

        (m_0,    b_uldm,    c_uldm) = ( (1.6*10**10) * (m22)**(-4/3),    1.1,    2.2)

    As for the concentration relative to CDM, there hasa not been a detailed study to date. Hoever, since the formation
    history and collapse time determines concentration, we can reasonably use a WDM concentration-mass relation
    such as the Lovell 2020 formula,
    i.e. simply c_wdm = c_uldm, with (c_scale, c_power) = (15, -0.3).
    or the Bose et al. formula with (c_scale, c_power) = (60, -0.17) and a very negligible redshift dependence.

    The default model was computed using the formalism presented by Schneider et al. (2015) using a ULDM power spectrum,
    and results in a sharper cutoff with (c_scale, c_power, c_power_inner) = (3.348, -0.489, 1.5460)

    The form for the ULDM concentration-mass relation turnover is (1 + c_scale * (mhm/m)^c_power_inner)^c_power

    Furthermore, Schive et al. 2014 (https://arxiv.org/pdf/1407.7762.pdf) found a redshift-dependent minimum ULDM halo mass
    given by

        M_min(z) = a^(-3/4) * (Zeta(z)/Zeta(0))^(1/4) * M_min,0     [solar masses]

    where a is the cosmic scale factor,

        M_min,0 = 4.4*10^7 * m22^(-3/2)     [solar masses]

    and

        Zeta(z) = (18*pi^2 + 82(Om(z) - 1) - 39(Om(z) - 1)^2) / Om(z),

    where Om(z) is the matter density parameter.

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param log10_m_uldm: log base 10 of the ULDM particle mass
    :param log10_fluc_amplitude: log base 10 of the fluctuation amplitude in the host halo density profile
    :param fluctuation_size_scale: sets the size of the fluctuations in host halo in units of de-Broglie wavelength
    :param fluctuation_size_dispersion: sets the variance of the size of the fluctuations
    :param n_fluc_scale: rescales the number of fluctuations
    :param velocity_scale: the velocity scale of the host halo used to convert particle mass to de-Broglie wavelength
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param mass_function_model_subhalos: mass function model for subhalos, see mass_function_models.py for a list
    :param kwargs_mass_function_subhalos: keyword arguments for the mass function model
    :param mass_function_model_fieldhalos: mass function model for field halos, see mass_function_models.py for a list
    :param kwargs_mass_function_fieldhalos: keyword arguments for the mass function model
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology, see Cosmology/cosmology.py
    :param uldm_plaw: sets the exponent for the core size/halo mass relation in ULDM
    :param flucs: bool; turns of/off fluctuations in the host halo density profile
    :param flucs_shape: the geometry in which to render fluctuations, options include 'aperture', 'ring', 'ellipse'
    :param flucs_args: keyword arguments for the specified geometry, see realization_extensions
    :param n_cut: maximum number of fluctuations to render; amplitudes of individual fluctuations get suppressed by
    sqrt(n_cut / n_total) if n_total > n_cut, where n_total is the number of expected fluctuations given the de-Broglie
    wavelength and the area in which the fluctuations area rendered
    :param r_ein: The characteristic angular used to convert fluctuations in convergence to fluctuations in mass
    :return: a realization of halos and fluctuations modeled as Gaussians in ULDM
    """

    # constants
    m22 = 10**(log10_m_uldm + 22)
    log_m0 = np.log10(1.6e10 * m22**(-4/3))

    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)

    #compute M_min as described in documentation
    log_m_min = MinHaloMassULDM(log10_m_uldm, pyhalo.astropy_cosmo, log_mlow)
    kwargs_density_profile = {}
    kwargs_density_profile['log10_m_uldm'] = log10_m_uldm
    kwargs_density_profile['scale_nfw'] = False
    kwargs_density_profile['uldm_plaw'] = uldm_plaw
    kwargs_wdm = {'z_lens': z_lens, 'z_source': z_source, 'log_mc': log_m0, 'sigma_sub': sigma_sub,
                  'log_mlow': log_m_min, 'log_mhigh': log_mhigh,
                  'mass_function_model_subhalos': mass_function_model_subhalos,
                  'kwargs_mass_function_subhalos': kwargs_mass_function_subhalos,
                  'mass_function_model_fieldhalos': mass_function_model_fieldhalos,
                  'kwargs_mass_function_fieldhalos': kwargs_mass_function_fieldhalos,
                  'concentration_model_subhalos': concentration_model_subhalos,
                  'kwargs_concentration_model_subhalos': kwargs_concentration_model_subhalos,
                  'concentration_model_fieldhalos': concentration_model_fieldhalos,
                  'kwargs_concentration_model_fieldhalos': kwargs_concentration_model_fieldhalos,
                  'truncation_model_subhalos': truncation_model_subhalos,
                  'kwargs_truncation_model_subhalos': kwargs_truncation_model_subhalos,
                  'truncation_model_fieldhalos': truncation_model_fieldhalos,
                  'kwargs_truncation_model_fieldhalos': kwargs_truncation_model_fieldhalos,
                  'shmf_log_slope': shmf_log_slope, 'cone_opening_angle_arcsec': cone_opening_angle_arcsec,
                  'log_m_host': log_m_host, 'r_tidal': r_tidal, 'LOS_normalization': LOS_normalization,
                  'geometry_type': geometry_type, 'kwargs_cosmo': kwargs_cosmo,
                  'mdef_subhalos': 'ULDM', 'mdef_field_halos': 'ULDM',
                  'kwargs_density_profile': kwargs_density_profile
                  }

    uldm_no_fluctuations = WDM(**kwargs_wdm)

    if flucs: # add fluctuations to realization
        ext = RealizationExtensions(uldm_no_fluctuations)
        lambda_dB = de_broglie_wavelength(log10_m_uldm, velocity_scale) # de Broglie wavelength in kpc

        if flucs_args=={}:
            raise Exception('Must specify fluctuation arguments, see realization_extensions.add_ULDM_fluctuations')

        a_fluc = 10 ** log10_fluc_amplitude
        m_psi = 10 ** log10_m_uldm

        zlens_ref, zsource_ref = 0.5, 2.0
        mhost_ref = 10**13.3
        rein_ref = 1.0
        r_perp_ref = rein_ref * uldm_no_fluctuations.lens_cosmo.cosmo.kpc_proper_per_asec(zlens_ref)

        model, _ = preset_concentration_models('DIEMERJOYCE19')
        concentration_model_for_host = model(pyhalo.astropy_cosmo)
        sigma_crit_ref = uldm_no_fluctuations.lens_cosmo.get_sigma_crit_lensing(zlens_ref, zsource_ref)
        c_host_ref = concentration_model_for_host.nfw_concentration(mhost_ref, z_lens)
        rhos_ref, rs_ref, _ = uldm_no_fluctuations.lens_cosmo.NFW_params_physical(mhost_ref, c_host_ref, zlens_ref)
        xref = r_perp_ref/rs_ref
        if xref < 1:
            Fxref = np.arctanh(np.sqrt(1 - xref ** 2)) / np.sqrt(1 - xref ** 2)
        else:
            Fxref = np.arctan(np.sqrt(-1 + xref ** 2)) / np.sqrt(-1 + xref ** 2)
        sigma_host_ref = 2 * rhos_ref * rs_ref * (1-Fxref)/(xref**2 - 1)

        r_perp = r_ein * uldm_no_fluctuations.lens_cosmo.cosmo.kpc_proper_per_asec(z_lens)
        sigma_crit = uldm_no_fluctuations.lens_cosmo.get_sigma_crit_lensing(z_lens, z_source)
        c_host = concentration_model_for_host.nfw_concentration(10**log_m_host, z_lens)
        rhos, rs, _ = uldm_no_fluctuations.lens_cosmo.NFW_params_physical(10**log_m_host, c_host, z_lens)
        x = r_perp / rs
        if x < 1:
            Fx = np.arctanh(np.sqrt(1 - x ** 2)) / np.sqrt(1 - x ** 2)
        else:
            Fx = np.arctan(np.sqrt(-1 + x ** 2)) / np.sqrt(-1 + x ** 2)
        sigma_host = 2 * rhos * rs * (1 - Fx) / (x ** 2 - 1)

        fluctuation_amplitude = a_fluc * (m_psi / 1e-22) ** -0.5 * \
                                (sigma_crit_ref/sigma_crit) * (sigma_host/sigma_host_ref)

        uldm_realization = ext.add_ULDM_fluctuations(de_Broglie_wavelength=lambda_dB,
                                fluctuation_amplitude=fluctuation_amplitude,
                                fluctuation_size=lambda_dB * fluctuation_size_scale,
                                fluctuation_size_variance=lambda_dB * fluctuation_size_scale * fluctuation_size_dispersion,
                                n_fluc_scale=n_fluc_scale,
                                shape=flucs_shape,
                                args=flucs_args,
                                n_cut=n_cut)
        return uldm_realization
    else:
        return uldm_no_fluctuations

def SIDM_core_collapse(z_lens, z_source, mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos, kwargs_sub_function=None,
        kwargs_field_function=None, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_ROCHE_GILMAN2020', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None, collapsed_halo_profile='SPL_CORE',
        kwargs_collapsed_profile={'x_core_halo': 0.05, 'x_match': 3.0, 'log_slope_halo': 3.0}):

    """
    Generates realizations of SIDM given the fraction of core-collapsed halos as a function of halo mass

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param mass_ranges_subhalos: a list of mass ranges for subhalo collapse, e.g.
    mass_ranges_subhalos = [[6, 8], [8, 9], ...]
    :param mass_ranges_field_halos: a list of mass ranges for subhalo collapse, e.g.
    mass_ranges_field_halos = [[6, 8], [8, 9], ...]
    :param probabilities_subhalos: a list of functions, or a list of the fraction of collapsed subhalos in each mass bin
    :param probabilities_field_halos: a list of functions, or a list of the fraction of collapsed field halos in each
    mass bin
    :param kwargs_sub_function: if probabilities_subhalos is a list of functions, specifies keyword arguments
    for each function
    :param kwargs_field_function: if probabilities_field_halos is a list of functions, specifies keyword arguments
    for each function
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param LOS_normalization: rescales the amplitude of the line-of-sight mass function
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology, see Cosmology/cosmology.py
    :param collapsed_halo_profile: string that sets the density profile of core-collapsed halos
    currently implemented models are SPL_CORE and GNFW (see example notebook)
    :param kwargs_collapsed_profile: keyword arguments for the collapsed profile (see example notebook)
    :return: a realization of dark matter structure in SIDM
    """

    two_halo_contribution = True
    delta_power_law_index = 0.0
    cdm = CDM(z_lens, z_source, sigma_sub, log_mlow, log_mhigh,
        concentration_model_subhalos, kwargs_concentration_model_subhalos,
        concentration_model_fieldhalos, kwargs_concentration_model_fieldhalos,
        truncation_model_subhalos, kwargs_truncation_model_subhalos,
        truncation_model_fieldhalos, kwargs_truncation_model_fieldhalos,
        shmf_log_slope, cone_opening_angle_arcsec, log_m_host,  r_tidal,
        LOS_normalization, two_halo_contribution, delta_power_law_index,
        geometry_type, kwargs_cosmo)

    extension = RealizationExtensions(cdm)
    index_collapsed = extension.core_collapse_by_mass(mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos, kwargs_sub_function, kwargs_field_function)
    sidm = extension.add_core_collapsed_halos(index_collapsed, collapsed_halo_profile, **kwargs_collapsed_profile)
    return sidm

def WDM_mixed(z_lens, z_source, log_mc, mixed_DM_frac, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        mass_function_model_subhalos='SHMF_MIXED_WDM_TURNOVER', kwargs_mass_function_subhalos={},
        mass_function_model_fieldhalos='MIXED_WDM_TURNOVER', kwargs_mass_function_fieldhalos={},
        concentration_model_subhalos='BOSE2016', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='BOSE2016', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_ROCHE_GILMAN2020', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        kwargs_density_profile={}):

    """
    Implements the mixed dark matter model presented by Keely et al. (2023)
    https://arxiv.org/pdf/2301.07265.pdf

    :param z_lens:
    :param z_source:
    :param log_mc:
    :param mixed_DM_frac: fraction of dark matter in CDM component
    :param sigma_sub:
    :param log_mlow:
    :param log_mhigh:
    :param mass_function_model_subhalos:
    :param kwargs_mass_function_subhalos:
    :param mass_function_model_fieldhalos:
    :param kwargs_mass_function_fieldhalos:
    :param concentration_model_subhalos:
    :param kwargs_concentration_model_subhalos:
    :param concentration_model_fieldhalos:
    :param kwargs_concentration_model_fieldhalos:
    :param truncation_model_subhalos:
    :param kwargs_truncation_model_subhalos:
    :param truncation_model_fieldhalos:
    :param kwargs_truncation_model_fieldhalos:
    :param shmf_log_slope:
    :param cone_opening_angle_arcsec:
    :param log_m_host:
    :param r_tidal:
    :param LOS_normalization:
    :param geometry_type:
    :param kwargs_cosmo:
    :param kwargs_density_profile:
    :return:
    """
    params = ['a_wdm', 'b_wdm', 'c_wdm']
    values_keeley_2023 = [0.5, 0.8, -3.0]
    for i, param in enumerate(params):
        if param not in kwargs_mass_function_subhalos.keys():
            kwargs_mass_function_subhalos[param] = values_keeley_2023[i]
        if param not in kwargs_mass_function_fieldhalos.keys():
            kwargs_mass_function_fieldhalos[param] = values_keeley_2023[i]
    kwargs_mass_function_subhalos['mixed_DM_frac'] = mixed_DM_frac
    kwargs_mass_function_fieldhalos['mixed_DM_frac'] = mixed_DM_frac
    kwargs_wdm = {'z_lens': z_lens, 'z_source': z_source, 'log_mc': log_mc, 'sigma_sub': sigma_sub,
                  'log_mlow': log_mlow, 'log_mhigh': log_mhigh,
                  'mass_function_model_subhalos': mass_function_model_subhalos,
                  'kwargs_mass_function_subhalos': kwargs_mass_function_subhalos,
                  'mass_function_model_fieldhalos': mass_function_model_fieldhalos,
                  'kwargs_mass_function_fieldhalos': kwargs_mass_function_fieldhalos,
                  'concentration_model_subhalos': concentration_model_subhalos,
                  'kwargs_concentration_model_subhalos': kwargs_concentration_model_subhalos,
                  'concentration_model_fieldhalos': concentration_model_fieldhalos,
                  'kwargs_concentration_model_fieldhalos': kwargs_concentration_model_fieldhalos,
                  'truncation_model_subhalos': truncation_model_subhalos,
                  'kwargs_truncation_model_subhalos': kwargs_truncation_model_subhalos,
                  'truncation_model_fieldhalos': truncation_model_fieldhalos,
                  'kwargs_truncation_model_fieldhalos': kwargs_truncation_model_fieldhalos,
                  'shmf_log_slope': shmf_log_slope, 'cone_opening_angle_arcsec': cone_opening_angle_arcsec,
                  'log_m_host': log_m_host, 'r_tidal': r_tidal, 'LOS_normalization': LOS_normalization,
                  'geometry_type': geometry_type, 'kwargs_cosmo': kwargs_cosmo,
                  'kwargs_density_profile': kwargs_density_profile
                  }
    return WDM(**kwargs_wdm)

def WDMGeneral(z_lens, z_source, log_mc, dlogT_dlogk, sigma_sub=0.025, log_mlow=6., log_mhigh=10.,
        truncation_model_subhalos='TRUNCATION_MEAN_DENSITY', kwargs_truncation_model_subhalos={},
        truncation_model_fieldhalos='SPLASHBACK', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3, r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        mdef_subhalos='TNFW', mdef_field_halos='TNFW', kwargs_density_profile={}):

    """
    This preset model implements a generalized treatment of warm dark matter, or any theory that produces a cutoff in
    the linear matter power spectrum. One additional free parameter is added to parameterize the cutoff in the matter
     power spectrum, dlogT_dlogk, which is the absolute value of the logarithmic derivative of the transfer function
     evaluated at k_1/2, the wavenumber corresponding to the half-mode mass (or log_mc).

     In this model, halo concentrations are computed following the formalism suggeesteed by Schneider et al. (2012) and
     Ludlow et al. (2016), in which the central density of a halo more or less is fully determined by the formation time.

     The mass function is computed from dlogT_dlogk based on the prescription by Stucker et al. (2021)
     https://arxiv.org/pdf/2109.09760.pdf

    :param z_lens: the lens redshift
    :param z_source: source redshift
    :param log_mc: the log (base 10) of the half-mode mass
    :param dlogT_dlogk: the absolute value of the logarithmic derivative of the transfer function at k_1/2; the model
    is calibrated for values between ~1 and ~3
    :param sigma_sub: amplitude of the subhalo mass function
    :param log_mlow: minimum halo mass to render
    :param log_mhigh: maximum halo mass to render
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param shmf_log_slope: logarithmic slope of the subhalo mass function
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume
    :param log_m_host: the log (base 10) of the host halo mass
    :param r_tidal: the core size in units of the scale radius of the host halo; subhalos are rendered uniformly in 3D
    inside this radius
    :param LOS_normalization: the amplitude of the LOS mass function relative to Sheth-Tormen
    :param geometry_type: CONE, DOUBLE_CONE, CYLINDER - sets the geometry of the rendering volume
    :param kwargs_cosmo: keyword arguments to set cosmology
    :param mdef_subhalos: mass definition for subhalos
    :param mdef_field_halos: mass definition for field halos
    :param kwargs_density_profile: keyword arguments for the density profile
    :return:
    """
    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source, kwargs_cosmo)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    # SET THE SPATIAL DISTRIBUTION MODELS FOR SUBHALOS AND FIELD HALOS:
    subhalo_spatial_distribution = ProjectedNFW
    fieldhalo_spatial_distribution = LensConeUniform

    kwargs_model_dlogT_dlogk = {'dlogT_dlogk': dlogT_dlogk}
    mass_function_model_subhalos, kwargs_mfunc_subs = preset_mass_function_models('STUCKER_SHMF', kwargs_model_dlogT_dlogk)
    mass_function_model_fieldhalos, kwargs_mfunc_field = preset_mass_function_models('STUCKER', kwargs_model_dlogT_dlogk)

    # SET THE TRUNCATION RADIUS FOR SUBHALOS AND FIELD HALOS
    model_subhalos, kwargs_truncation_model_subhalos = truncation_models(truncation_model_subhalos,
                                                                         kwargs_truncation_model_subhalos)
    kwargs_truncation_model_subhalos['lens_cosmo'] = pyhalo.lens_cosmo
    truncation_model_subhalos = model_subhalos(**kwargs_truncation_model_subhalos)

    model_fieldhalos, kwargs_truncation_model_fieldhalos = truncation_models(truncation_model_fieldhalos,
                                                                             kwargs_truncation_model_fieldhalos)
    kwargs_truncation_model_fieldhalos['lens_cosmo'] = pyhalo.lens_cosmo
    truncation_model_fieldhalos = model_fieldhalos(**kwargs_truncation_model_fieldhalos)

    # SET THE CONCENTRATION-MASS RELATION FOR SUBHALOS AND FIELD HALOS
    concentration_model = 'FROM_FORMATION_HISTORY'
    model_subhalos, kwargs_concentration_model_subhalos = preset_concentration_models(concentration_model,
                                                                                      kwargs_model_dlogT_dlogk)
    concentration_model_CDM = preset_concentration_models('DIEMERJOYCE19')[0]
    kwargs_concentration_model_subhalos['concentration_cdm_class'] = concentration_model_CDM
    kwargs_concentration_model_subhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_subhalos['log_mc'] = log_mc
    concentration_model_subhalos = model_subhalos(**kwargs_concentration_model_subhalos)

    model_fieldhalos, kwargs_concentration_model_fieldhalos = preset_concentration_models(concentration_model,
                                                                                          kwargs_model_dlogT_dlogk)
    kwargs_concentration_model_fieldhalos['concentration_cdm_class'] = concentration_model_CDM
    kwargs_concentration_model_fieldhalos['cosmo'] = pyhalo.astropy_cosmo
    kwargs_concentration_model_fieldhalos['log_mc'] = log_mc
    concentration_model_fieldhalos = model_fieldhalos(**kwargs_concentration_model_fieldhalos)

    # NOW THAT THE CLASSES ARE SPECIFIED, WE SORT THE KEYWORD ARGUMENTS AND CLASSES INTO LISTS
    population_model_list = ['SUBHALOS', 'LINE_OF_SIGHT', 'TWO_HALO']

    mass_function_class_list = [mass_function_model_subhalos,
                                mass_function_model_fieldhalos,
                                mass_function_model_fieldhalos]
    kwargs_mfunc_subs.update({'log_mlow': log_mlow,
                       'log_mhigh': log_mhigh,
                       'm_pivot': 10 ** 8,
                       'power_law_index': shmf_log_slope,
                       'log_m_host': log_m_host,
                       'sigma_sub': sigma_sub,
                       'delta_power_law_index': 0.0,
                        'log_mc': log_mc})
    kwargs_mfunc_field.update({'log_mlow': log_mlow,
                  'log_mhigh': log_mhigh,
                  'LOS_normalization': LOS_normalization,
                  'm_pivot': 10 ** 8,
                  'log_m_host': log_m_host,
                  'delta_power_law_index': 0.0,
                  'log_mc': log_mc})
    kwargs_mass_function_list = [kwargs_mfunc_subs, kwargs_mfunc_field, kwargs_mfunc_field]
    spatial_distribution_class_list = [subhalo_spatial_distribution, fieldhalo_spatial_distribution, fieldhalo_spatial_distribution]
    kwargs_subhalos_spatial = {'m_host': 10 ** log_m_host, 'zlens': z_lens,
                               'rmax2d_arcsec': cone_opening_angle_arcsec / 2, 'r_core_units_rs': r_tidal,
                               'lens_cosmo': pyhalo.lens_cosmo}
    kwargs_los_spatial = {'cone_opening_angle': cone_opening_angle_arcsec, 'geometry': geometry}
    kwargs_spatial_distribution_list = [kwargs_subhalos_spatial, kwargs_los_spatial, kwargs_los_spatial]

    kwargs_halo_model = {'truncation_model_subhalos': truncation_model_subhalos,
                         'concentration_model_subhalos': concentration_model_subhalos,
                         'truncation_model_field_halos': truncation_model_fieldhalos,
                         'concentration_model_field_halos': concentration_model_fieldhalos,
                         'kwargs_density_profile': kwargs_density_profile}

    realization_list = pyhalo.render(population_model_list, mass_function_class_list, kwargs_mass_function_list,
                                     spatial_distribution_class_list, kwargs_spatial_distribution_list,
                                     geometry, mdef_subhalos, mdef_field_halos, kwargs_halo_model, nrealizations=1)
    return realization_list[0]

def CDMFromEmulator(z_lens, z_source, emulator_input, kwargs_cdm):
    """
    This generates a realization of subhalos using an emulator of the semi-analytic modeling code Galacticus, and
     generates line-of-sight halos from a mass function parameterized as Sheth-Tormen.

    :param z_lens: main deflector redshift
    :param z_source: sourcee redshift
    :param emulator_input: either an array or a callable function

    if callable: a function that returns an array of
    1) subhalo masses at infall [M_sun]
    2) subhalo projected x position [kpc]
    3) subhalo projected y position [kpc]
    4) subhalo final_bound_mass [M_sun]
    5) subhalo concentrations at infall

    if not callable: an array with shape (N_subhalos 5) that contains masses, positions x, positions y, etc.

    Mass convention is m200 with respect to the critical density of the Universe at redshift Z_infall, where Z_infall is
    the infall redshift (not necessarily the redshift at the time of lensing).

    :param cone_opening_angle_arcsec: the opening angle of the double cone rendering volume in arcsec
    :param log_mlow: log10(minimum halo mass) rendered, or a function that returns log_mlow given a redshift
    :param log_mhigh: log10(maximum halo mass) rendered, or a function that returns log_mlow given a redshift
    :param LOS_normalization: rescaling of the line of sight halo mass function relative to Sheth-Tormen
    :param log_m_host: log10 host halo mass in M_sun
    :param kwargs_other: allows for additional keyword arguments to be specified when creating realization

    The following optional keywords specify a concentration-mass relation for field halos parameterized as a power law
    in peak height. If they are not set in the function call, pyHalo assumes a default concentration-mass relation from Diemer&Joyce
    :param c0: amplitude of the mass-concentration relation at 10^8
    :param log10c0: logarithmic amplitude of the mass-concentration relation at 10^8 (only if c0_mcrelation is None)
    :param beta: logarithmic slope of the mass-concentration-relation pivoting around 10^8
    :param zeta: modifies the redshift evolution of the mass-concentration-relation
    :param two_halo_contribution: whether to include the two-halo term for correlated structure near the main deflector
    :param kwargs_halo_mass_function: keyword arguments passed to the LensingMassFunction class
    (see Cosmology.lensing_mass_function)
    :return: a realization of CDM halos
    """

    # we create a realization of only line-of-sight halos by setting sigma_sub = 0.0
    kwargs_cdm['sigma_sub'] = 0.0
    cdm_halos_LOS = CDM(z_lens, z_source, **kwargs_cdm)
    # get lens_cosmo class from class containing LOS objects; note that this will work even if there are no LOS halos
    lens_cosmo = cdm_halos_LOS.lens_cosmo

    # now create subhalos from the specified properties using the TNFWSubhaloEmulator class
    halo_list = []
    if callable(emulator_input):
        subhalo_infall_masses, subhalo_x_kpc, subhalo_y_kpc, subhalo_final_bound_masses, \
        subhalo_infall_concentrations = emulator_input()
    else:
        subhalo_infall_masses = emulator_input[:, 0]
        subhalo_x_kpc = emulator_input[:, 1]
        subhalo_y_kpc = emulator_input[:, 2]
        subhalo_final_bound_masses = emulator_input[:, 3]
        subhalo_infall_concentrations = emulator_input[:, 4]

    for i in range(0, len(subhalo_infall_masses)):
        halo = TNFWSubhaloEmulator(subhalo_infall_masses[i],
                                   subhalo_x_kpc[i],
                                   subhalo_y_kpc[i],
                                   subhalo_final_bound_masses[i],
                                   subhalo_infall_concentrations[i],
                                   z_lens, lens_cosmo)
        halo_list.append(halo)

    # combine the subhalos with line-of-sight halos
    subhalos_from_emulator = Realization.from_halos(halo_list, lens_cosmo, kwargs_halo_model={},
                                                    msheet_correction=False, rendering_classes=None)
    return cdm_halos_LOS.join(subhalos_from_emulator)




from copy import deepcopy
import numpy as np


class DeltaFunction(object):
    """
    This class generates masses from a delta function normalized with respect to a
    background density, a mass, and a volume

    number of objects = density * volume / mass
    """
    name = 'DELTA_FUNCTION'
    def __init__(self, mass, volume, rho, draw_poisson, *args, **kwargs):
        """

        :param mass: mass of objects to render
        :param volume: rendering volume
        :param rho: a density
        :param draw_poisson: whether or not to draw from a poisson distribution
        """
        self.volume = volume
        self.mass = mass
        self.rho = rho
        self.draw_poisson = draw_poisson
        self.n_mean = self.rho * self.volume / self.mass
        self.first_moment = self.mass

    @classmethod
    def from_redshift(cls, z, delta_z, geometry_class, kwargs_model):
        """
        Creates the class from arbitrary functions that specify the amplitude and slope of the
         halo mass function at different redshifts

        """
        astropy = geometry_class.cosmo.astropy
        density_to_MsunperMpc = geometry_class.cosmo.density_to_MsunperMpc
        rho_dm_crit = astropy.Odm(0.) * astropy.critical_density(z).value * density_to_MsunperMpc
        volume = geometry_class.volume_element_comoving(z, delta_z)
        rho = kwargs_model['LOS_normalization'] * kwargs_model['mass_fraction'] * rho_dm_crit
        return DeltaFunction(kwargs_model['mass'], volume, rho, kwargs_model['draw_poisson'])

    def draw(self):

        """
        :return: an array of masses
        """

        if self.draw_poisson:
            n = int(np.random.poisson(self.n_mean))
        else:
            n = int(np.round(self.n_mean))

        if n > 0:
            return np.array([self.mass] * n)
        else:
            return np.array([])



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.tnfw import TNFW as TNFWLenstronomy
import numpy as np

class TNFWFieldHalo(Halo):

    """
    The base class for a truncated NFW halo
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        mdef = 'TNFW'
        super(TNFWFieldHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                           lens_cosmo_instance, args, unique_tag)

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['TNFW']

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        if not hasattr(self, '_params_physical'):

            [concentration, rt] = self.profile_args
            rhos, rs, r200 = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z_eval)
            self._params_physical = {'rhos': rhos * self._rescale_norm, 'rs': rs, 'r200': r200, 'r_trunc_kpc': rt}

        return self._params_physical

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):

            [concentration, rt] = self.profile_args
            Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle(self.mass, concentration, self.z)

            x, y = np.round(self.x, 4), np.round(self.y, 4)

            Rs_angle = np.round(Rs_angle, 10)
            theta_Rs = np.round(theta_Rs, 10)
            r_trunc_arcsec = rt / self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)

            kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                      'center_x': x, 'center_y': y, 'r_trunc': r_trunc_arcsec}]

            self._kwargs_lenstronomy = kwargs

        return self._kwargs_lenstronomy, None

    @property
    def z_eval(self):
        """
        Returns the halo redshift
        """
        return self.z

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            truncation_radius_kpc = self._truncation_class.truncation_radius_halo(self)
            self._profile_args = (self.c, truncation_radius_kpc)

        return self._profile_args

    @property
    def bound_mass(self):
        """
        Computes the mass inside the virial radius (with truncation effects included)
        :return: the mass inside r = c * r_s
        """
        prof = TNFWLenstronomy()
        kwargs_profile = self.lenstronomy_params[0][0]
        alpha_rs = kwargs_profile['alpha_Rs']
        rs = kwargs_profile['Rs']
        r_trunc = kwargs_profile['r_trunc']
        r = self.c * rs
        rho0 = prof.alpha2rho0(alpha_rs, rs)
        mass_3d = prof.mass_3d(r, rs, rho0, r_trunc)
        mass_3d_infall = prof.mass_3d(r, rs, rho0, 1000*rs)
        return (mass_3d / mass_3d_infall) * self.mass


class TNFWSubhalo(TNFWFieldHalo):
    """
    Defines a truncated NFW halo that is a subhalo of the host dark matter halo
    """

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                self._zeval = self.z
            else:
                self._zeval = self.z_infall

        return self._zeval

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_params_physical'):
            [concentration, rt] = self.profile_args
            rhos, rs, r200 = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z_eval)
            self._params_physical = {'rhos': rhos, 'rs': rs, 'r200': r200, 'r_trunc_kpc': rt}

        return self._params_physical



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.splcore import SPLCORE
import numpy as np

class PowerLawSubhalo(Halo):

    """
    The base class for a halo modeled as a power law profile
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._prof = SPLCORE()
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        mdef = 'SPL_CORE'
        super(PowerLawSubhalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                              lens_cosmo_instance, args, unique_tag)

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        if not hasattr(self, '_params_physical'):
            (concentration, gamma, x_core_halo) = self.profile_args
            rhos, rs, _ = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)

            if 'x_match' in self._args.keys():
                if self._args['x_match'] == 'c':
                    x_match = concentration
                else:
                    x_match = self._args['x_match']
            else:
                # r_vmax = 2.16 * rs
                x_match = 2.16

            r_match_arcsec = x_match * rs / kpc_per_arcsec
            fx = np.log(1 + x_match) - x_match / (1 + x_match)
            m = 4 * np.pi * rs ** 3 * rhos * fx
            r_core_arcsec = x_core_halo * r_match_arcsec / x_match

            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2

            rho0 = m / self._prof.mass_3d(r_match_arcsec, sigma_crit_arcsec, r_core_arcsec, gamma)
            # units 1 / arcsec

            rho0 *= sigma_crit_arcsec * kpc_per_arcsec ** -3
            self._params_physical = {'rho0': rho0, 'r_core': x_core_halo * rs}

        return self._params_physical

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):

            (concentration, gamma, x_core_halo) = self.profile_args
            rhos, rs, r200 = self._lens_cosmo.NFW_params_physical(self.mass, concentration, self.z)
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)

            if 'x_match' in self._args.keys():
                if self._args['x_match'] == 'c':
                    x_match = concentration
                else:
                    x_match = self._args['x_match']
            else:
                # r_vmax = 2.16 * rs
                x_match = 2.16 # r_vmax

            r_match_arcsec = x_match * rs / kpc_per_arcsec
            fx = np.log(1+x_match) - x_match/(1 + x_match)
            m = 4 * np.pi * rs ** 3 * rhos * fx
            r_core_arcsec = x_core_halo * r_match_arcsec / x_match

            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2

            rho0 = m/self._prof.mass_3d(r_match_arcsec, sigma_crit_arcsec, r_core_arcsec, gamma)
            sigma0 = rho0 * r_core_arcsec

            self._lenstronomy_args = [{'sigma0': sigma0, 'gamma': gamma, 'center_x': self.x, 'center_y': self.y,
                                      'r_core': r_core_arcsec}]

        return self._lenstronomy_args, None

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['SPL_CORE']

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                self._zeval = self.z
            else:
                self._zeval = self.z_infall

        return self._zeval

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            concentration = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
            gamma = self._args['log_slope_halo']
            x_core_halo = self._args['x_core_halo']
            self._profile_args = (concentration, gamma, x_core_halo)

        return self._profile_args


class PowerLawFieldHalo(PowerLawSubhalo):
    """
    Class that defines a power law halo in the field
    """
    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            concentration = self._concentration_class.nfw_concentration(self.mass, self.z)
            gamma = self._args['log_slope_halo']
            x_core_halo = self._args['x_core_halo']
            self._profile_args = (concentration, gamma, x_core_halo)

        return self._profile_args

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evaluate the concentration-mass relation
        """
        return self.z



import pytest
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
import numpy.testing as npt
from pyHalo.Halos.tidal_truncation import TruncationRN, TruncationRoche, TruncationSplashBack, TruncateMeanDensity
from pyHalo.truncation_models import truncation_models
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
import os


class TestTruncation(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy)
        self.lenscosmo = LensCosmo(0.5, 1.5, cosmo)

    def test_load_models(self):

        model_name_list = ['TRUNCATION_R50', 'TRUNCATION_RN', 'TRUNCATION_ROCHE', 'TRUNCATION_ROCHE_GILMAN2020',
                           'SPLASHBACK', 'TRUNCATION_MEAN_DENSITY']
        kwargs_model_list = [{}, {'LOS_truncation_factor': 50.}, {'RocheNorm': 1.0, 'm_power': 1./3, 'RocheNu': 2.0/3.0}, {},
                             {}, {}]
        for model,kwargs in zip(model_name_list, kwargs_model_list):
            mod, kw = truncation_models(model)
            kwargs.update(kw)
            kwargs['lens_cosmo'] = self.lenscosmo
            _ = mod(**kwargs)

    def test_truncation_RN(self):

        N = 200
        halo_mass = 10 ** 8
        truncation_RN = TruncationRN(self.lenscosmo, N)
        r200_kpc = truncation_RN.truncation_radius(halo_mass, 0.5)
        r200_kpc_true = self.lenscosmo.NFW_params_physical(halo_mass, 16.0, 0.5)[-1]
        npt.assert_almost_equal(r200_kpc, r200_kpc_true)

    def test_truncation_roche(self):

        norm = 1.4
        m_power = 1. / 3
        nu = 2. / 3
        r3d_subhalo = 65.0
        halo_mass = 10 ** 8
        truncation_roche = TruncationRoche(None, norm, m_power, nu)
        r200_kpc = truncation_roche.truncation_radius(halo_mass, r3d_subhalo)
        r200_kpc_true = norm * (halo_mass/10**7) ** m_power * (r3d_subhalo/50)**nu
        npt.assert_almost_equal(r200_kpc, r200_kpc_true, 3)

    def test_truncation_splashback(self):

        class DummyHalo(object):
            def __init__(self, m, z):
                self.mass = m
                self.z = z
        truncation_splashback = TruncationSplashBack(self.lenscosmo)
        rt = truncation_splashback.truncation_radius(10**8, 0.4)
        halo = DummyHalo(10**8, 0.4)
        rt_halo = truncation_splashback.truncation_radius_halo(halo)
        npt.assert_almost_equal(rt, rt_halo)

    def test_truncation_mean_density(self):

        class DummyHalo(object):
            def __init__(self, m, z):
                self.mass = m
                self.z = z
                self.z_eval = z
                self.rperi_units_r200 = 0.7
                self.c = 16.0

        halo = DummyHalo(10 ** 8, 0.4)
        median_rt_over_rs = 2.0
        c_power = 4.0
        cmodel = ConcentrationDiemerJoyce(self.lenscosmo.cosmo, scatter=False)
        c_theory = cmodel.nfw_concentration(10 ** 8, 0.4)
        truncation = TruncateMeanDensity(self.lenscosmo, median_rt_over_rs, c_power)
        r_t = truncation.truncation_radius(10**8, 0.4, c_theory, 16.0, 0.7)
        r_t_halo = truncation.truncation_radius_halo(halo)
        npt.assert_almost_equal(r_t, r_t_halo)

        rt_over_rs_theory = median_rt_over_rs * (16.0 / c_theory) ** c_power * (0.7 / 0.5)
        rs = self.lenscosmo.NFW_params_physical(10**8, 16.0, 0.4)[1]
        npt.assert_almost_equal(rs * rt_over_rs_theory, r_t)


if __name__ == '__main__':
    pytest.main()



from pyHalo.single_realization import Realization
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Rendering.halo_population import HaloPopulation
from pyHalo.utilities import generate_lens_plane_redshifts
from pyHalo.Halos.lens_cosmo import LensCosmo
from astropy.cosmology import FlatLambdaCDM
from pyHalo.defaults import cosmo_default

class pyHalo(object):

    """
    The main class used for generating realizations (see example notebook)
    """

    def __init__(self, zlens, zsource, kwargs_cosmo=None):

        """

        :param zlens:
        :param zsource:
        :param kwargs_cosmo:
        """
        if kwargs_cosmo is None:
            astropy_instance = None
            kwargs_cosmo = cosmo_default.cosmo_param_dictionary
        else:
            # required keys for custom seteup:  'H0', 'Ob0', 'Om0', 'sigma8', 'flat', 'ns', 'power_law'
            astropy_instance = FlatLambdaCDM(H0=kwargs_cosmo['H0'],
                                             Ob0=kwargs_cosmo['Ob0'],
                                             Om0=kwargs_cosmo['Om0'])
        self.zlens = zlens
        self.zsource = zsource
        self.cosmology = Cosmology(astropy_instance, kwargs_cosmo)
        self.lens_cosmo = LensCosmo(self.zlens, self.zsource, self.cosmology)
        self.colossus_cosmo = self.cosmology.colossus

    @property
    def astropy_cosmo(self):
        return self.cosmology.astropy

    def render(self, population_model_list, mass_function_class_list, kwargs_mass_function_list,
                   spatial_distribution_class_list, kwargs_spatial_distribution_list,
               geometry_class, mdef_subhalos, mdef_field_halos, kwargs_halo_model, nrealizations=1):

        """

        :param population_model_list:
        :param mass_function_class_list:
        :param kwargs_mass_function_list:
        :param spatial_distribution_class_list:
        :param kwargs_spatial_distribution_list:
        :param geometry_class:
        :param mdef_subhalos:
        :param mdef_field_halos:
        :param kwargs_halo_model:
        :param nrealizations:
        :return:
        """
        realization_list = []
        for i in range(0, nrealizations):
            masses, x_arcsec, y_arcsec, r3d, redshifts, subhalo_flag, rendering_classes = self.render_masses_positions(population_model_list,
                                mass_function_class_list,
                                kwargs_mass_function_list,
                                spatial_distribution_class_list,
                                kwargs_spatial_distribution_list,
                                geometry_class)
            realization_list.append(self.create_realization(masses, x_arcsec, y_arcsec, r3d, redshifts, subhalo_flag, rendering_classes,
                                                            geometry_class, mdef_subhalos, mdef_field_halos, kwargs_halo_model))
        return realization_list

    def render_masses_positions(self, population_model_list,
                                mass_function_class_list,
                                kwargs_mass_function,
                                spatial_distribution_class_list,
                                kwargs_spatial_distribution,
                                geometry_class):

        """

        :param population_model_list:
        :param mass_function_class_list:
        :param kwargs_mass_function:
        :param spatial_distribution_class_list:
        :param kwargs_spatial_distribution:
        :param geometry_class:
        :return:
        """

        plane_redshifts, redshift_spacing = generate_lens_plane_redshifts(self.zlens, self.zsource)
        population_model = HaloPopulation(population_model_list,
                                              mass_function_class_list,
                                              kwargs_mass_function,
                                              spatial_distribution_class_list,
                                              kwargs_spatial_distribution,
                                              self.lens_cosmo,
                                              geometry_class,
                                              plane_redshifts,
                                              redshift_spacing)
        masses, x_arcsec, y_arcsec, r3d, redshifts, subhalo_flag = population_model.render()
        return masses, x_arcsec, y_arcsec, r3d, redshifts, subhalo_flag, population_model.rendering_classes

    def create_realization(self, masses, x_arcsec, y_arcsec, r3d, redshifts, subhalo_flag,
                           rendering_classes, geometry_class, mdef_subhalos, mdef_field_halos, kwargs_halo_model):
        """

        :param masses:
        :param x_arcsec:
        :param y_arcsec:
        :param r3d:
        :param redshifts:
        :param subhalo_flag:
        :param rendering_classes:
        :param geometry_class:
        :param mdef_subhalos:
        :param mdef_field_halos:
        :param convergence_sheet_correction:
        :return:
        """

        mdefs = []
        for i in range(0, len(masses)):
            if subhalo_flag[i]:
                mdefs += [mdef_field_halos]
            else:
                mdefs += [mdef_subhalos]
        realization = Realization(masses, x_arcsec, y_arcsec, r3d, mdefs, redshifts, subhalo_flag, self.lens_cosmo,
                                  kwargs_halo_model=kwargs_halo_model,
                                  mass_sheet_correction=True,
                                  rendering_classes=rendering_classes, geometry=geometry_class)
        return realization



from copy import deepcopy
import numpy as np
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw, WDMPowerLaw, MixedWDMPowerLaw
from colossus.lss.mass_function import massFunction

__all__ = ['ShethTormen', 'ShethTormenMixedWDM', 'ShethTormenTurnover']


class ShethTormen(CDMPowerLaw):
    """
    This class samples from the Sheth-Tormen halo mass function

    kwargs_model for this class include:
    1) log_mlow: minimum halo mass to render (log base 10)
    2) log_mhigh: maximum halo mass to render (log base 10)
    3) m_pivot: the pivot mass; the logarithmic slope is defined around the pivot
    4) LOS_normalization: rescales the ammplitude of the mass function at the pivot scale
    5) delta_power_law_index: adjusts the logarithmic slope of the mass function around the pivot scale
    """
    name = 'SHETH_TORMEN'
    @classmethod
    def from_redshift(cls, z, delta_z, geometry_class, kwargs_model):
        """

        :param z:
        :param delta_z:
        :param geometry_class:
        :param kwargs_model:
        :return:
        """
        _ = geometry_class.cosmo.colossus
        m_pivot = kwargs_model['m_pivot']
        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 10)
        m_h = m * h
        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / m_pivot), np.log10(dndM_comoving), 1)
        plaw_index = coeffs[0] + kwargs_model['delta_power_law_index']
        norm_dv = 10 ** coeffs[1] / (m_pivot**plaw_index)
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = kwargs_model['LOS_normalization'] * norm_dv * volume_element_comoving
        return ShethTormen(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                           kwargs_model['draw_poisson'], normalization)

class ShethTormenTurnover(WDMPowerLaw):
    """
    This class samples from the Sheth-Tormen halo mass function with a break at a scale 10^log_mc

    kwargs_model for this class include:
    1) log_mlow: minimum halo mass to render (log base 10)
    2) log_mhigh: maximum halo mass to render (log base 10)
    3) m_pivot: the pivot mass; the logarithmic slope is defined around the pivot
    4) LOS_normalization: rescales the ammplitude of the mass function at the pivot scale
    5) delta_power_law_index: adjusts the logarithmic slope of the mass function around the pivot scale
    6) log_mc: log 10 of the break scale
    7) a_wdm: shifts the position of the cutoff (see expression below)
    8) b_wdm: the inner exponent of the cutoff, determines the sharpness of the cut (see expression below)
    9) c_wdm: the outer exponent of the cutoff, determines the logarithmic slope below log_mc together with b_wdm

    cutoff has the functional form (1 + a_wdm * (m_c/m) ^ b_wdm) ^ c_wdm
    where m_c = 10^log_mc
    """
    name = 'SHETH_TORMEN'
    @classmethod
    def from_redshift(cls, z, delta_z, geometry_class, kwargs_model):
        """

        :param z:
        :param delta_z:
        :param geometry_class:
        :param kwargs_model:
        :return:
        """
        _ = geometry_class.cosmo.colossus
        m_pivot = kwargs_model['m_pivot']
        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 10)
        m_h = m * h
        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / m_pivot), np.log10(dndM_comoving), 1)
        plaw_index = coeffs[0] + kwargs_model['delta_power_law_index']
        norm_dv = 10 ** coeffs[1] / (m_pivot ** plaw_index)
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = kwargs_model['LOS_normalization'] * norm_dv * volume_element_comoving
        return ShethTormenTurnover(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                           kwargs_model['draw_poisson'], normalization, kwargs_model['log_mc'],
                                   kwargs_model['a_wdm'], kwargs_model['b_wdm'], kwargs_model['c_wdm'])

class ShethTormenMixedWDM(MixedWDMPowerLaw):
    """
    This class samples from the Sheth-Tormen halo mass function with a break at a scale 10^log_mc for mixed warm and
    cold dark matter

    kwargs_model for this class include:
    1) log_mlow: minimum halo mass to render (log base 10)
    2) log_mhigh: maximum halo mass to render (log base 10)
    3) m_pivot: the pivot mass; the logarithmic slope is defined around the pivot
    Note that this only has affect when delta_power_law_index != 0
    4) log_mc: log 10 of the break scale
    5) a_wdm: shifts the position of the cutoff (see expression below)
    6) b_wdm: the inner exponent of the cutoff, determines the sharpness of the cut (see expression below)
    7) c_wdm: the outer exponent of the cutoff, determines the logarithmic slope below log_mc together with b_wdm

    cutoff has the functional form (1 + a_wdm * (m_c/m) ^ b_wdm) ^ c_wdm
    where m_c = 10^log_mc
    """
    name = 'SHETH_TORMEN'
    @classmethod
    def from_redshift(cls, z, delta_z, geometry_class, kwargs_model):
        """

        :param z:
        :param delta_z:
        :param geometry_class:
        :param kwargs_model:
        :return:
        """
        colossus_cosmo = geometry_class.cosmo.colossus
        m_pivot = kwargs_model['m_pivot']
        h = geometry_class.cosmo.h
        # To M_sun / h units
        m = np.logspace(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], 10)
        m_h = m * h
        dndlogM = massFunction(m_h, z, q_out='dndlnM', model='sheth99')
        dndM_comoving_h = dndlogM / m
        # three factors of h for the (Mpc/h)^-3 to Mpc^-3 conversion
        dndM_comoving = dndM_comoving_h * h ** 3
        coeffs = np.polyfit(np.log10(m / m_pivot), np.log10(dndM_comoving), 1)
        plaw_index = coeffs[0] + kwargs_model['delta_power_law_index']
        norm_dv = 10 ** coeffs[1]
        volume_element_comoving = geometry_class.volume_element_comoving(z, delta_z)
        normalization = kwargs_model['LOS_normalization'] * norm_dv * volume_element_comoving / (m_pivot ** plaw_index)
        return ShethTormenMixedWDM(kwargs_model['log_mlow'], kwargs_model['log_mhigh'], plaw_index,
                                         kwargs_model['draw_poisson'], normalization, kwargs_model['log_mc'],
                                   kwargs_model['a_wdm'], kwargs_model['b_wdm'],
                                   kwargs_model['c_wdm'], kwargs_model['mixed_DM_frac'])




import numpy as np
from pyHalo.Rendering.SpatialDistributions.base import SpatialDistributionBase

class Correlated2D(SpatialDistributionBase):
    """
    This class generates points from an arbitrary 2D probability distribution
    """
    def __init__(self, geometry, smooth_scale=1.):

        """
        :param geometry: an instance of Geometry
        :param smooth_scale: a smoothing scale that removes the regular grid
        pattern in the rendered points
        """
        self._geo = geometry
        self._smooth_scale = smooth_scale
        super(Correlated2D, self).__init__()

    def draw(self, n, r_max, density, z_plane, shift_x=0., shift_y=0.):

        """

        :param n: the number of points to draw
        :param r_max: the radius in arcsec of the rendering area, should correspond to
        the angular size of density
        :param density: the 2D probability density to sample from
        :param z_plane: the redshift of the lens plane
        :param shift_x: moves the center of rendered points
        from (0, 0) to (shift_x, shift_y)
        :param shift_y: moves the center of rendered points
        from (0, 0) to (shift_x, shift_y)
        :return: x and y samples
        """
        norm = np.sum(density)
        if norm == 0:
            raise Exception('2D probability distribution not normalizable')

        density = density / np.sum(density)

        s = density.shape[0]
        p = density.reshape(-1)

        values = np.arange(len(p))
        pairs = np.indices(dimensions=(s, s)).T

        x_coordinates_arcsec = np.linspace(-r_max, r_max, s)
        y_coordinates_arcsec = np.linspace(-r_max, r_max, s)

        inds = np.random.choice(values, p=p, size=n, replace=True)
        locations = pairs.reshape(-1, 2)[inds]
        x_sample_pixel, y_sample_pixel = locations[:, 0], locations[:, 1]

        x_sample_arcsec = x_coordinates_arcsec[x_sample_pixel]
        y_sample_arcsec = y_coordinates_arcsec[y_sample_pixel]

        smoothing = self._smooth_scale * r_max / s
        x_sample_arcsec += np.random.normal(0., smoothing, len(x_sample_arcsec))
        y_sample_arcsec += np.random.normal(0., smoothing, len(y_sample_arcsec))

        kpc_per_asec = self._geo.kpc_per_arcsec(z_plane)

        x_sample_arcsec += shift_x
        y_sample_arcsec += shift_y

        x_kpc, y_kpc = x_sample_arcsec * kpc_per_asec, y_sample_arcsec * kpc_per_asec

        return x_kpc, y_kpc



import numpy.testing as npt
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Halos.lens_cosmo import LensCosmo
import numpy as np
import pytest
from astropy.cosmology import FlatLambdaCDM
from astropy.constants import G, c
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
import astropy.units as un
from pyHalo.Halos.HaloModels.NFW import NFWSubhhalo, NFWFieldHalo
from pyHalo.Cosmology.cosmology import Cosmology
from lenstronomy.Cosmo.lens_cosmo import LensCosmo as LensCosmoLenstronomy
from pyHalo.Halos.HaloModels.generalized_nfw import GeneralNFWSubhalo

class TestLensCosmo(object):

    def setup_method(self):

        kwargs_cosmo = {'Om0': 0.2}
        self.cosmo = Cosmology(cosmo_kwargs=kwargs_cosmo)
        zlens, zsource = 0.3, 1.7
        self.lens_cosmo = LensCosmo(zlens, zsource, self.cosmo)
        self.h = self.cosmo.h

    def test_const(self):

        D_ds = self.cosmo.D_A(0.3, 1.7)
        D_d = self.cosmo.D_A_z(0.3)
        D_s = self.cosmo.D_A_z(1.7)

        c_Mpc_sec = un.Quantity(c, unit=un.Mpc / un.s)
        G_Mpc3_Msun_sec2 = un.Quantity(G, unit=un.Mpc ** 3 / un.s ** 2 / un.solMass)

        const = c_Mpc_sec ** 2 / (4 * np.pi * G_Mpc3_Msun_sec2)
        sigma_crit_mpc = const.value * D_s / (D_d * D_ds)
        sigma_crit_kpc = sigma_crit_mpc * 1000 ** -2

        npt.assert_almost_equal(self.lens_cosmo.sigma_crit_lensing/sigma_crit_mpc, 1, 4)
        npt.assert_almost_equal(self.lens_cosmo.sigma_crit_lens_kpc/sigma_crit_kpc, 1, 4)

    def test_sigma_crit_mass(self):

        area = 2.
        sigma_crit_mass = self.lens_cosmo.sigma_crit_mass(0.7, area)
        sigma_crit = self.lens_cosmo.get_sigma_crit_lensing(0.7, 1.7)
        npt.assert_almost_equal(sigma_crit_mass, sigma_crit * area)

    def test_subhalo_accretion(self):

        zi = [self.lens_cosmo.z_accreted_from_zlens(10**8, 0.5)
              for _ in range(0, 20000)]

        h, b = np.histogram(zi, bins=np.linspace(0.5, 6, 20))

        # number at the lens redshift should be about 5x that at redshift 4
        ratio = h[0]/h[12]
        npt.assert_almost_equal(ratio/5 - 1, 0., 1)

    def test_mhm_convert(self):

        mthermal = 5.3
        mhm = self.lens_cosmo.mthermal_to_halfmode(mthermal)
        mthermal_out = self.lens_cosmo.halfmode_to_thermal(mhm)
        npt.assert_almost_equal(mthermal/mthermal_out, 1, 2)

        fsl = self.lens_cosmo.mhm_to_fsl(10**8.)
        npt.assert_array_less(fsl, 100)

    def test_nfw_definitions_wrt_lenstronomy(self):

        astropy = FlatLambdaCDM(70, 0.3)
        cosmo = Cosmology(astropy)
        zlens = 0.5
        zsource = 2.0
        lc = LensCosmoLenstronomy(zlens, zsource, astropy)
        lens_cosmo = LensCosmo(zlens, zsource, cosmo)

        rho0_lenstronomy, rs_lenstronomy, r200_lenstronomy = lc.nfwParam_physical(10 ** 8, 16.0)
        rho0, rs, r200 = lens_cosmo.nfwParam_physical(10 ** 8, 16.0, zlens)
        npt.assert_almost_equal(rho0_lenstronomy/rho0, 1, 4)
        npt.assert_almost_equal(rs/rs_lenstronomy, 1.0, 4)
        npt.assert_almost_equal(r200/r200_lenstronomy, 1.0, 4)

        rho0_kpc, rs_kpc, r200_kpc = lens_cosmo.NFW_params_physical(10 ** 8, 16.0, zlens)
        npt.assert_almost_equal(rho0_kpc, rho0*1000**-3)
        npt.assert_almost_equal(rs_kpc, rs*1000)
        npt.assert_almost_equal(r200_kpc, r200 * 1000)


if __name__ == '__main__':
    pytest.main()





import numpy as np
from scipy.optimize import root

def stucker_suppression_params(dlogT_dlogk, gamma=5.0):

    """
    Maps from the logarithmic derivative of the transfer function at the half-mode scale to the a, b, c parameters
    that describe the suppression of the halo mass function.

    Credit for these routines goes to Jens Stucker: https://arxiv.org/pdf/2109.09760.pdf
    Please cite the paper when using this routine
    :param dlogT_dlogk: the logarithmic derivative of the transfer function at k half-mode
    :return: the a, b, c parameters for the mass function suppression
    """

    fabg = (-1.0 + 0.5 ** (-1 / gamma))
    beta = dlogT_dlogk * (1 + fabg) / fabg / gamma

    if (beta > 6.) | (beta < 1.5):
        print("beta = %.2f is outside the validated range [1.5 ... 6]!"
                      " Usage at your own risk!" % beta)
    mu = np.array((0.2651, 1.638, 16.51))
    nu = np.array((0.3656, -0.0994, -0.9466))
    (m20, m50, m80) = 1.0*mu*beta**nu
    a_stucker, b, c = mscales_to_abc(m20, m50, m80)
    a = a_stucker ** b
    return a, b, c

def mscales_to_abc(m20, m50, m80):
    """Uniquely maps the parameters m20, m50, m80 to a, b, c
    Credit for these routines goes to Jens Stucker: https://arxiv.org/pdf/2109.09760.pdf
    Please cite the paper when using this routine
    """
    def _equations(par):
        a, b, c =  10**par[0], par[1], par[2]
        ms = _supp_mscale(a, b, c, frac=np.array([0.2,0.5,0.8]))
        eq = abs(np.log10(ms) - np.log10([m20, m50, m80]))/0.01
        return eq
    p0 = np.array([np.log10(m50), 2., -1.])
    res = root(_equations, p0, method="lm")
    return 10**res.x[0], res.x[1], res.x[2]

def _supp_mscale(a, b, c, frac=0.5):
    """The mass scale where the surpression reaches a given value, given a,b,c
    a, b, c : parameters (a >= 0, b >= 0, c <= 0)
    frac : desired surpression factor
    returns : the mass M where f(M)=frac, units are as the parameter "a"

    Credit for these routines goes to Jens Stucker: https://arxiv.org/pdf/2109.09760.pdf
    Please cite the paper when using this routine
    """
    if (a < 0) | (b < 0) | (c > 0):
        # This is just here, to tell the fitting function when it goes out of bounds
        return np.nan
    return a / (frac**(1./c) - 1.)**(1./b)



from pyHalo.Halos.halo_base import Halo
import numpy as np

class NFWFieldHalo(Halo):

    """
    The main class for an NFW field halo profile without truncation

    See the base class in Halos/halo_base.py for the required routines for any instance of a Halo class
    """

    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):

        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        mdef = 'NFW'
        super(NFWFieldHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """

        return self.z

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['NFW']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        (concentration) = self.profile_args
        Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle(self.mass, concentration, self.z)

        x, y = np.round(self.x, 4), np.round(self.y, 4)
        Rs_angle = np.round(Rs_angle, 10)
        theta_Rs = np.round(theta_Rs, 10)

        kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                  'center_x': x, 'center_y': y}]

        return kwargs, None

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            concentration = self.c
            self._profile_args = (concentration)

        return self._profile_args

class NFWSubhhalo(NFWFieldHalo):

    """
    The main class for an NFW subhalo profile without truncation

    See the base class in Halos/halo_base.py for the required routines for any instance of a Halo class
    """

    @property
    def z_eval(self):
        """
        Returns the redshift at which to evalate the concentration-mass relation
        """
        if not hasattr(self, '_zeval'):

            if 'evaluate_mc_at_zlens' in self._args.keys() and self._args['evaluate_mc_at_zlens']:
                self._zeval = self.z
            else:
                self._zeval = self.z_infall

        return self._zeval



import numpy as np
import pickle
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from scipy.interpolate import RegularGridInterpolator
from pyHalo.Halos.util import tau_mf_interpolation
from colossus.lss import peaks
from colossus.halo import splashback


class TruncationSplashBack(object):

    def __init__(self, lens_cosmo):
        """
        This computes the splashback radius of a halo as the truncation radius (appropriate for field halos)
        See Diemer (2020)
        :param lens_cosmo: an instance of LensCosmo
        """
        self._lens_cosmo = lens_cosmo

    def truncation_radius_halo(self, halo):
        """
        Thiis method computes the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        return self.truncation_radius(halo.mass, halo.z)

    def truncation_radius(self, halo_mass, z):
        """
        Computes the radius r_N of an NFW halo
        :param halo_mass: halo mass (m200 with respect to critical density at z)
        :param z: redshift
        :param halo_concentration: the halo concentration
        :param lens_cosmo: an instance of LensCosmo
        :return: the truncation radius
        """
        h = self._lens_cosmo.cosmo.h
        nu200m = peaks.peakHeight(halo_mass * h, z)
        RspR200m_units_rvir, _ = splashback.splashbackModel('RspR200m', nu200m=nu200m, z=z, rspdef='sp-apr-mn')
        r200m_mpc = self._lens_cosmo.rN_M(halo_mass * h, z, 200.0) / self._lens_cosmo.cosmo.astropy.Om0 ** (1./3)
        rt_kpc = RspR200m_units_rvir * r200m_mpc * 1e3
        return rt_kpc

class TruncationRN(object):

    def __init__(self, lens_cosmo, LOS_truncation_factor=50):
        """
        This implements a tidal truncation at r_N, where N is some overdensity with respect to the critical density of
        the Universe at z
        :param lens_cosmo: an instance of LensCosmo
        :param LOS_truncation_factor: the multiple of the overdensity threshold at which to truncte the halo, e.g. N=200
        would truncate at r200
        """
        self._lens_cosmo = lens_cosmo
        self._N = LOS_truncation_factor

    def truncation_radius_halo(self, halo):
        """
        Thiis method computes the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        return self.truncation_radius(halo.mass, halo.z)

    def truncation_radius(self, halo_mass, z):
        """
        Computes the radius r_N of an NFW halo
        :param halo_mass: halo mass (m200 with respect to critical density at z)
        :param z: redshift
        :param lens_cosmo: an instance of LensCosmo
        :return: the truncation radius
        """
        h = self._lens_cosmo.cosmo.h
        rN_physical_mpc = self._lens_cosmo.rN_M(halo_mass * h, z, self._N)
        return rN_physical_mpc*1000

class TruncationRoche(object):

    def __init__(self, lens_cosmo=None, RocheNorm=1.4, m_power=1./3, RocheNu=2./3):
        """
        This implements a tidal truncation for subhalos of the form

        r_t = norm * (m / 10^7) ^ m_power * (r3d/50 kpc)^r3d_power [kpc]

        The default values were calibrated for a subhalo inside an isothermal host potential
        https://ui.adsabs.harvard.edu/abs/2016PhRvD..94d3505C/abstract

        :param norm: the overall scale
        :param m_power: exponent for the dependence on halo mass
        :param r3d_power: exponent for the dependence on 3D position inside host
        """
        self._norm = RocheNorm
        self._m_power = m_power
        self._r3d_power = RocheNu

    def truncation_radius_halo(self, halo):

        """
        Thiis method computess the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        return self.truncation_radius(halo.mass, halo.r3d)

    def truncation_radius(self, subhalo_mass, subhalo_r3d):

        """
        :param M: m200
        :param r3d: 3d radial position in the halo (physical kpc)
        :return: the truncation radius in physical kpc
        """
        m_units_7 = subhalo_mass / 10 ** 7
        radius_units_50 = subhalo_r3d / 50
        rtrunc_kpc = self._norm * m_units_7 ** self._m_power * radius_units_50 ** self._r3d_power
        return np.round(rtrunc_kpc, 3)

class AdiabaticTidesTruncation(object):
    """
    An example of the type of class we want to create and implement in pyHalo
    """

    def __init__(self, lens_cosmo, log_m_host, z_host, mass_loss_interp):
        """

        :param lens_cosmo: an instacee of the LensCosmo class
        :param log_m_host: the host halo mass
        :param z_host: the redshift of the host halo
        :param mass_loss_interp: an instance of RegularGridInterpolator
        The interpolator should take as input a point
        (log10(concentration), log10(r_pericenter), c_host) and return the
        asymptotic final bound mass divided by the infall mass
        """
        self._mass_loss_interp = mass_loss_interp
        min_max_c = [1.0, 10 ** 2.7]
        min_max_rperi = [10 ** -2.0, 10 ** 0.2]
        self._lens_cosmo = lens_cosmo
        cmodel = ConcentrationDiemerJoyce(self._lens_cosmo.cosmo.astropy, scatter=False)
        c_host = cmodel.nfw_concentration(10**log_m_host, z_host)
        self._chost = c_host
        self._host_dynamical_time = self._lens_cosmo.halo_dynamical_time(10**log_m_host, z_host, c_host)
        self._min_c = min_max_c[0]
        self._max_c = min_max_c[1]
        self._min_rperi = min_max_rperi[0]
        self._max_rperi = min_max_rperi[1]
        self._tau_mf_interpolation = tau_mf_interpolation()

    def truncation_radius_halo(self, halo):
        """
        This function solves for the truncation radius divided by the halo scale radius (tau) given an instance of
        Halo (see pyhalo.HaloModels.TNFW) and other arguments for the interpolation function

        :param halo: an instance of the Halo class (should be TNFW)

        :return tau: the truncation radius divided by the halo's scale radius
        """

        # compute the halo parameeters
        c = halo.c
        # now make sure that the points are inside the region where we computed the interpolation
        r_pericenter_over_r200 = np.absolute(halo.rperi_units_r200)
        r_pericenter_over_r200 = max(self._min_rperi, r_pericenter_over_r200)
        r_pericenter_over_r200 = min(self._max_rperi, r_pericenter_over_r200)
        c = max(self._min_c, c)
        c = min(self._max_c, c)
        point = (np.log10(c), np.log10(r_pericenter_over_r200), self._chost)
        # evaluate the mass loss
        log10mass_loss_fraction_asymptotic = float(self._mass_loss_interp(point))

        time_since_infall = halo.time_since_infall
        n_orbits = time_since_infall / self._host_dynamical_time
        mass_loss_fraction = self._temporal_mass_loss(10 ** log10mass_loss_fraction_asymptotic, n_orbits)
        log10mass_loss_fraction = np.log10(mass_loss_fraction)

        # solve for tau
        log10c = np.log10(c)
        point = (log10c, log10mass_loss_fraction)
        point = self._make_params_in_bounds_tau_evaluate(point)
        log10tau = float(self._tau_mf_interpolation(point))
        tau = 10 ** log10tau
        _, rs, _ = self._lens_cosmo.NFW_params_physical(halo.mass, halo.c, halo.z_eval)
        return tau * rs

    @staticmethod
    def _temporal_mass_loss(mass_loss_asymptotic, n_orbits):
        """
        This routine interpolates between zero mass loss at infall and the asymptotic value
        predicted by the adiabatic tides model
        """
        mass_loss = mass_loss_asymptotic * (1 / mass_loss_asymptotic) ** (1. / (1. + 0.6 * n_orbits))
        return mass_loss

    def _make_params_in_bounds_tau_evaluate(self, point):
        """
        This routine makes sure the arguments for the initerpolation are inside the domain of the function.
        """
        (log10c, log10mass_loss_fraction) = point
        log10c = max(self._min_c, log10c)
        log10c = min(self._max_c, log10c)
        log10mass_loss_fraction = max(-1.5, log10mass_loss_fraction)
        log10mass_loss_fraction = min(-0.01, log10mass_loss_fraction)
        return (log10c, log10mass_loss_fraction)

class TruncateMeanDensity(object):

    def __init__(self, lens_cosmo, median_rt_over_rs=2.0, c_power=3.0):
        """

        :param lens_cosmo:
        :param median_rt_over_rs:
        :param c_power:
        """

        self._norm = median_rt_over_rs
        self._cpower = c_power
        self.lens_cosmo = lens_cosmo
        self._concentration_cdm = ConcentrationDiemerJoyce(lens_cosmo.cosmo,
                                                           scatter=False)

    def truncation_radius_halo(self, halo):

        """
        Thiis method computess the truncation radius using the class attributes of an instance of Halo
        :param halo: an instance of halo
        :return: the truncation radius in physical kpc
        """
        c_median = self._concentration_cdm.nfw_concentration(halo.mass, halo.z_eval)
        c_actual = halo.c
        halo_rpericenter = halo.rperi_units_r200
        return self.truncation_radius(halo.mass, halo.z, c_median, c_actual, halo_rpericenter)

    def truncation_radius(self, halo_mass, halo_redshift, c_median, c_actual, r_peri):
        """

        :param halo_mass:
        :param halo_redshift:
        :param c_median:
        :param c_actual:
        :param r_peri:
        :return:
        """
        rt_over_rs = self._norm * (c_actual / c_median) ** self._cpower * (r_peri / 0.5)
        _, rs, _ = self.lens_cosmo.NFW_params_physical(halo_mass, c_actual, halo_redshift)
        return rs * rt_over_rs



from pyHalo.Rendering.SpatialDistributions.uniform import Uniform, LensConeUniform
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.Cosmology.cosmology import Cosmology
import numpy as np
import numpy.testing as npt
import pytest

class TestUniform(object):

    def setup_method(self):

        cosmo = Cosmology()
        self.geometry = Geometry(cosmo, 0.5, 1.5, 6., 'DOUBLE_CONE')
        self.uni = Uniform(3, self.geometry)
        self.lenscone_uni = LensConeUniform(6., self.geometry)

    def test_uniform(self):

        x, y = self.uni.draw(1000, 0.3, 1., 0., 0.)
        kpc_per_asec = self.geometry.kpc_per_arcsec(0.3)
        x *= kpc_per_asec ** -1
        y *= kpc_per_asec ** -1
        r2d = np.hypot(x, y)
        npt.assert_equal(True, max(r2d) < 3)

        x, y = self.uni.draw(1000, 0.3, 0.5, 0., 0.)
        x *= kpc_per_asec ** -1
        y *= kpc_per_asec ** -1
        r2d = np.hypot(x, y)
        npt.assert_equal(True, max(r2d) < 1.5)

    def test_lens_cone_uniform(self):

        x, y = self.lenscone_uni.draw(10000, 0.5)
        kpc_per_asec = self.geometry.kpc_per_arcsec(0.5)
        x *= kpc_per_asec ** -1
        y *= kpc_per_asec ** -1
        x2, y2 = self.lenscone_uni.draw(10000, 0.9)
        kpc_per_asec = self.geometry.kpc_per_arcsec(0.9)
        x2 *= kpc_per_asec ** -1
        y2 *= kpc_per_asec ** -1

        scale1 = self.geometry.rendering_scale(0.5)
        scale2 = self.geometry.rendering_scale(0.9)

        max1 = max(np.hypot(x, y))
        max2 = max(np.hypot(x2, y2))
        npt.assert_almost_equal(max1/max2, scale1/scale2, 2)

    def test_distribution(self):

        x, y = self.uni.draw(1000000, 0.2, 1., 0., 0.)
        r2 = np.hypot(x, y)
        rbins = np.arange(1., 3.+0.5, 0.5)
        n = []
        for i in range(0, len(rbins)-1):
            condition = np.logical_and(r2 >= rbins[i], r2<rbins[i+1])
            N = np.sum(condition)
            area = np.pi * (rbins[i+1]**2 - rbins[i]**2)
            n.append(N/area)
        npt.assert_almost_equal(n[0]/n[1], 1, 1)
        npt.assert_almost_equal(n[2]/n[3], 1, 1)

if __name__ == '__main__':

    pytest.main()





class SpatialDistributionBase(object):

    @classmethod
    def from_Mhost(cls, *args, **kwargs):
        raise Exception('The spatial distribution model specified for subhalos must have a class method .from_Mhost')

    def draw(cls, *args, **kwargs):
        raise Exception('The spatial distribution model specified for subhalos must have a method draw()')



import pytest
import numpy.testing as npt
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tidal_truncation import AdiabaticTidesTruncation
from pyHalo.Halos.HaloModels.TNFW import TNFWSubhalo
from pyHalo.Halos.lens_cosmo import LensCosmo
import numpy as np

class DummyInterp(object):

    def __call__(self, x):
        #np.log10(c), np.log10(r_pericenter_over_r200), self._log10_galaxy_rs, self._log10_galaxy_m
        (log10c, log10_rperi_r200, chost) = x
        rperi_term = (10**log10_rperi_r200 / 1.0) ** 2
        c_term = (10**log10c / 500) **0.5
        return np.log10(c_term * rperi_term)

class TestAdiabaticTides(object):

    def setup_method(self):
        self.lens_cosmo_instance = LensCosmo(0.5, 1.5)
        self.att = AdiabaticTidesTruncation(self.lens_cosmo_instance,
                                            13.3, 0.5, mass_loss_interp=DummyInterp())


    def test_rt(self):

        mass = 10 ** 8.0
        x = 0.1
        y = 0.5
        r3d = None
        z = 0.5
        sub_flag = True
        kwargs_density_profile = {'evaluate_mc_at_zlens': True}
        truncation_class = self.att
        concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo_instance.cosmo.astropy,
                                                       scatter=False)
        unique_tag = 1.0
        halo_1 = TNFWSubhalo(mass, x, y, r3d, z,
                 sub_flag, self.lens_cosmo_instance, kwargs_density_profile,
                 truncation_class, concentration_class, unique_tag)
        _ = halo_1.c
        halo_1._rperi_units_r200 = 0.4
        halo_1._time_since_infall = 1.0
        profile_params_1 = halo_1.lenstronomy_params[0]
        tau_1 = profile_params_1[0]['r_trunc']/profile_params_1[0]['Rs']

        halo_2 = TNFWSubhalo(mass, x, y, r3d, z,
                             sub_flag, self.lens_cosmo_instance, kwargs_density_profile,
                             truncation_class, concentration_class, unique_tag)
        _ = halo_2.c
        halo_2._rperi_units_r200 = 0.6
        halo_2._time_since_infall = 1.0
        profile_params_2 = halo_2.lenstronomy_params[0]
        tau_2 = profile_params_2[0]['r_trunc'] / profile_params_2[0]['Rs']

        npt.assert_equal(tau_2 > tau_1, True)

        halo_3 = TNFWSubhalo(mass, x, y, r3d, z,
                             sub_flag, self.lens_cosmo_instance, kwargs_density_profile,
                             truncation_class, concentration_class, unique_tag)
        _ = halo_3.c
        halo_3._c *= 5.0
        halo_3._rperi_units_r200 = 0.6
        halo_3._time_since_infall = 1.0
        profile_params_3 = halo_3.lenstronomy_params[0]
        tau_3 = profile_params_3[0]['r_trunc'] / profile_params_3[0]['Rs']
        npt.assert_equal(tau_3 > tau_2, True)

if __name__ == '__main__':
     pytest.main()




import pytest
import numpy as np
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Cosmology.geometry import Geometry
import numpy.testing as npt
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Rendering.SpatialDistributions.nfw import ProjectedNFW
from pyHalo.Rendering.MassFunctions.mass_function_base import CDMPowerLaw
from pyHalo.Rendering.subhalos import Subhalos, normalization_sigmasub

class TestSubhalos(object):

    def setup_method(self):

        sigma_sub = 0.75
        kwargs_mass_function = {'log_mlow': 6.0, 'log_mhigh': 8.0, 'm_pivot': 1e8, 'power_law_index': -1.9,
                                'delta_power_law_index': 0.054, 'draw_poisson': False, 'log_m_host': 13.5,
                                'sigma_sub': sigma_sub}
        zlens = 0.8
        zsource = 1.5
        cosmo = Cosmology()
        geometry = Geometry(cosmo, zlens, zsource, 6.0, 'DOUBLE_CONE')
        lens_cosmo = LensCosmo(zlens, zsource, cosmo)
        mhost = 10 ** kwargs_mass_function['log_m_host']
        spatial_distribution_model = ProjectedNFW.from_Mhost(mhost, zlens, geometry.cone_opening_angle/2,
                                                                           0.05, lens_cosmo)
        mass_function_model = CDMPowerLaw
        self.model = Subhalos(mass_function_model, kwargs_mass_function, spatial_distribution_model,
                              geometry, lens_cosmo)
        self.kwargs_mass_function = kwargs_mass_function
        self._norm_sigma_sub = normalization_sigmasub(kwargs_mass_function['sigma_sub'],
                                                      10**kwargs_mass_function['log_m_host'], zlens,
                                                      lens_cosmo.cosmo.kpc_proper_per_asec(zlens),
                                                      geometry.cone_opening_angle,
                                                      kwargs_mass_function['power_law_index']+kwargs_mass_function['delta_power_law_index'],
                                                      kwargs_mass_function['m_pivot'])

    def test_mass_rendered(self):

        model = CDMPowerLaw(self.kwargs_mass_function['log_mlow'], self.kwargs_mass_function['log_mhigh'],
                              self.kwargs_mass_function['power_law_index']+self.kwargs_mass_function['delta_power_law_index'],
                              False, self._norm_sigma_sub)
        mtheory = model.first_moment
        m_rendered = np.sum(self.model.render()[0])
        npt.assert_array_less(abs(1-mtheory/m_rendered), 0.03)

    def test_rendering(self):

        m, x, y, r3d_kpc, z, subhalo_flag = self.model.render()
        npt.assert_equal(len(m), len(x))
        npt.assert_equal(len(x), len(y))
        npt.assert_equal(len(r3d_kpc), len(z))
        npt.assert_equal(len(subhalo_flag), len(x))
        for flag in subhalo_flag:
            npt.assert_equal(True, flag)
        for zi in z:
            npt.assert_equal(zi, 0.8)

    def test_sheets(self):
        kwargs_out, profile_names_out, redshifts = self.model.convergence_sheet_correction(kappa_scale=1.0,
                                                                                           log_mlow=6.0,
                                                                                           log_mhigh=10.0)
        npt.assert_equal(len(kwargs_out), len(profile_names_out))
        npt.assert_equal(len(profile_names_out), len(redshifts))

if __name__ == '__main__':
    pytest.main()




import numpy.testing as npt
import pytest
from pyHalo.preset_models import CDM
from pyHalo.plotting_routines import plot_subhalo_bound_mass, \
    plot_subhalo_mass_functon, plot_concentration_mass_relation, plot_halo_mass_function

class TestPlottingRoutines(object):

    def setup_method(self):
        self.realization = CDM(0.5, 1.5, sigma_sub=0.025, LOS_normalization=0.0)

    def test_mass_function_plot(self):

        plot_subhalo_mass_functon(self.realization)
        plot_subhalo_mass_functon(self.realization, bound_mass_function=True)
        plot_halo_mass_function(self.realization, z_eval=0.5)
        plot_halo_mass_function(self.realization, z_eval=None)
        plot_halo_mass_function(self.realization, z_eval=[0.5, 0.6])

    def test_bound_mass_function_plot(self):

        plot_subhalo_bound_mass(self.realization)

    def test_plot_mc_relation(self):

        plot_concentration_mass_relation(self.realization, z_eval='z_lens')
        plot_concentration_mass_relation(self.realization, z_eval=0.5)
        plot_concentration_mass_relation(self.realization, z_eval=[0.4, 0.6])


if __name__ == '__main__':
     pytest.main()



import matplotlib.pyplot as plt
import numpy as np
from pyHalo.single_realization import realization_at_z

plt.rcParams['axes.linewidth'] = 2.5
plt.rcParams['xtick.major.width'] = 3.5
plt.rcParams['xtick.major.size'] = 10
plt.rcParams['xtick.minor.size'] = 6
plt.rcParams['ytick.minor.size'] = 6
plt.rcParams['ytick.major.width'] = 3.5
plt.rcParams['ytick.major.size'] = 10
plt.rcParams['ytick.labelsize'] = 20
plt.rcParams['xtick.labelsize'] = 20
plt.rcParams['font.family']


def plot_halo_mass_function(realization, z_eval=None, ax=None, color='k', kwargs_plot={},
                                     log_mlow=6.0, log_mhigh=10.0, nbins=25):

    """
    Makes a log-log plot of the halo mass function for a given realization.

    :param realization:
    :param z_eval:
    :param ax:
    :param color:
    :param kwargs_plot:
    :param log_mlow:
    :param log_mhigh:
    :param nbins:
    :param show_errorbars:
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    xlabel = 'halo mass '+r'$\left[M_{\odot}\right]$'
    ylabel = r'$n\left(m\right)$'
    if z_eval is None:
        masses = realization.masses
    elif isinstance(z_eval, float) or isinstance(z_eval, int):
        real = realization_at_z(realization, z_eval)[0]
        masses = real.masses
    elif isinstance(z_eval, list):
        masses = []
        for halo in realization.halos:
            if halo.is_subhalo:
                continue
            if halo.z >= z_eval[0] and halo.z < z_eval[1]:
                masses.append(halo.mass)
    else:
        raise Exception('z_eval must be one of: a specific redshift (i.e. a floating point number), or an'
                        'interval z_eval = [z_min, z_max]')
    h, m = np.histogram(masses, bins=np.logspace(log_mlow, log_mhigh, nbins))
    ax.plot(m[0:-1], h, color=color, **kwargs_plot)
    ax.set_xlabel(xlabel, fontsize=14)
    ax.set_ylabel(ylabel, fontsize=14)
    ax.set_xscale('log')
    ax.set_yscale('log')

def plot_concentration_mass_relation(realization, z_eval, ax=None, color='k', kwargs_plot={},
                                     log_mlow=6.0, log_mhigh=10.0, nbins=20, show_errorbars=False):

    """
    Makes a log-log plot of the concentration-mass relation for a given realization.

    :param realization: an instance of the Realization class
    :param z_eval: the redshift at which to evaluate the mc relation; can be one of 'z_lens', a flaot,
    or an interval [z_min, z_max]
    :param ax: figure axis
    :param color: the color for the lines in the figure
    :param kwargs_plot: keyword arguments passed to plot()
    :param log_mlow: minimum halo mass
    :param log_mhigh: maximum halo mass
    :param nbins: number of mass bins along the x axis
    :param show_errorbars: bool; show the logarithmic scatter
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)

    if z_eval == 'z_lens':
        xlabel = 'infall halo mass ' + r'$\left[M_{\odot}\right]$'
        ylabel = r'$c\left(m\right)$'
        halo_mass_list = []
        halo_concentration_list = []
        for halo in realization.halos:
            if halo.is_subhalo:
                halo_mass_list.append(halo.mass)
                halo_concentration_list.append(halo.c)
    elif isinstance(z_eval, float) or isinstance(z_eval, int):
        xlabel = 'halo mass ' + r'$\left[M_{\odot}\right]$'
        ylabel = r'$c\left(m\right)$'
        halo_mass_list = []
        halo_concentration_list = []
        for halo in realization.halos:
            if halo.is_subhalo and halo.z == z_eval:
                halo_mass_list.append(halo.mass)
                halo_concentration_list.append(halo.c)
    elif isinstance(z_eval, list):
        xlabel = 'halo mass ' + r'$\left[M_{\odot}\right]$'
        ylabel = r'$c\left(m\right)$'
        halo_mass_list = []
        halo_concentration_list = []
        for halo in realization.halos:
            if halo.z >= z_eval[0] and halo.z < z_eval[1]:
                halo_mass_list.append(halo.mass)
                halo_concentration_list.append(halo.c)
    else:
        raise Exception('z_eval must be one of: z_lens, a specific redshift (i.e. a floating point number), or an'
                        'interval z_eval = [z_min, z_max]')

    halo_masses = np.array(halo_mass_list)
    halo_concentrations = np.array(halo_concentration_list)

    x = np.logspace(log_mlow, log_mhigh, nbins + 1)
    c = []
    c_error = []
    for i in range(0, len(x) - 1):
        cond1 = halo_masses >= x[i]
        cond2 = halo_masses < x[i + 1]
        inds = np.where(np.logical_and(cond1, cond2))[0]
        c_median = np.median(halo_concentrations[inds])
        c_std = np.std(halo_concentrations[inds])
        c.append(c_median)
        c_error.append(c_std)

    if show_errorbars:
        ax.errorbar(x[0:-1], c, yerr=c_error, fmt='none', color=color)
    ax.plot(x[0:-1], c, color=color, **kwargs_plot)
    ax.set_xlabel(xlabel, fontsize=15)
    ax.set_ylabel(ylabel, fontsize=15)
    ax.set_xscale('log')
    ax.set_yscale('linear')


def plot_subhalo_bound_mass(realization, ax=None, color='k', kwargs_plot={},
                            log_mlow=6.0, log_mhigh=10.0):
    """
    Plots the bound mass of subhalos defined as the mass inside the infall virial radius of a truncated
    profile. Note: the halo mass definition must have a truncation radius specified for this method

    :param realization: an instance of Realization
    :param ax: figure axis
    :param color: color for the lines
    :param kwargs_plot: keyword arguments passed to plot
    :param log_mlow: minimum infall halo mass along x axis
    :param log_mhigh: maximum infall halo mass along x axis
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = plt.subplot(111)
    masses = []
    mass_bound = []
    for halo in realization.halos:
        if halo.is_subhalo:
            masses.append(halo.mass)
            mass_bound.append(halo.bound_mass)

    ax.scatter(masses, mass_bound, color=color, **kwargs_plot)
    ax.plot(masses, masses, color='0.5', lw=2, linestyle=':')
    ax.set_xlim(10**log_mlow, 10**log_mhigh)
    ax.set_xlabel('infall halo mass '+r'$\left[M_{\odot}\right]$', fontsize=15)
    ax.set_ylabel('subhalo mass '+r'$M\left(r < r_{200}\right)$', fontsize=15)
    ax.set_xscale('log')
    ax.set_yscale('log')

def plot_subhalo_mass_functon(realization, log_m_low=6.0, log_m_high=10.0, bound_mass_function=False, nbins=20,
                                  n_bootstrap=10, ax=None, color='k', kwargs_plot={}):
        """
        Makes a log-log plot of the infall mass versus the final bound mass of tidally stripped subhalos
        :param realization: an instance of Realization
        :param log_mlow: minimum infall halo mass along x axis
        :param log_mhigh: maximum infall halo mass along x axis
        :param bound_mass_function:
        :param nbins: number of bins along x axis
        :param n_bootstrap: number of bootstrap resampling iteratins to compute vertical error bars
        :param ax: figure axis
        :param color: color for the lines
        :param kwargs_plot: keyword arguments passed to plot
        :return:
        """
        if ax is None:
            fig = plt.figure()
            ax = plt.subplot(111)
        masses = []
        for halo in realization.halos:
            if halo.is_subhalo:
                if bound_mass_function:
                    masses.append(halo.bound_mass)
                else:
                    masses.append(halo.mass)
        h = np.empty((n_bootstrap, nbins))
        for i in range(0, n_bootstrap):
            inds = np.random.randint(0, len(masses), len(masses))
            _h, _x = np.histogram(np.log10(masses)[inds], bins=nbins, range=(log_m_low, log_m_high))
            h[i, :] = _h
        median = np.median(h, axis=0)
        standard_dev = np.std(h, axis=0)
        logm = _x[1:] - (_x[1] - _x[0])/2
        ax.errorbar(logm, median, yerr=standard_dev, fmt='none', color=color)
        ax.plot(logm, median, color=color, **kwargs_plot)
        ax.set_ylim(median[-1] / 2, median[0] * 2)
        ax.set_xlabel('infall halo mass ' + r'$\left[M_{\odot}\right]$', fontsize=15)
        ax.set_ylabel(r'$n\left(m\right)$', fontsize=15)
        ax.set_yscale('log')
